/*----------------------------------------------------------------------------*/
/*                                                                            */
/*                             CharSeqLongSet.cpp                             */
/*                                                                            */
/*----------------------------------------------------------------------------*/


#include "CharSeqLongSet.h"


namespace Text {

  long int nnewline(FILE* fp);
  int ncharLine(FILE* fp);
}


//
//
//  CONSTRUCTOR: CharSeq::CharSeqLongSet::CharSeqLongSet()
//
//  PURPOSE: reads from file the set of character sequences.
//
//  ARGUMENT:
//   ---> fp: position of the first character of the first sequence in the 
//            file. The sequences occupy consecutive lines in the file. Each of 
//            these lines contains only the sequence followed by the line 
//            terminator. After the last sequence's line, the file terminates.
//
//  NOTES:
//  1. the n-th sequence in the set is the n-th sequence in the file, for 
//     n=0,...,N-1.
//
//  2. When the function terminates, fp points to the end-of-file character.
//
//
CharSeq::CharSeqLongSet::CharSeqLongSet(FILE* fp) {

  const long int SET_START = ftell(fp);
  numseq = Text::nnewline(fp);

  fseek(fp, SET_START, SEEK_SET);
  seqlen = Text::ncharLine(fp);

  set = new char*[numseq];
  for(long int s=0; s<numseq; ++s)
    set[s] = new char[seqlen+1];

  // set[n] points to the n-th sequence in the set (n=0,...,N-1)
  fseek(fp, SET_START, SEEK_SET); 
  for(long int s=0; s<numseq; ++s)
    fscanf(fp, "%s ", set[s]);  
}


//
//
//  DESTRUCTOR: CharSeq::CharSeqLongSet::~CharSeqLongSet()
//
//  PURPOSE: releases the heap memory occupied by the set.
//
//  COMPLEXITY: O(numseq)
//
//
CharSeq::CharSeqLongSet::~CharSeqLongSet() {
 
  for(long int s=0; s<numseq; ++s)
    delete[] set[s];

  delete[] set;
}


//
//
//  FUNCTION: CharSeq::CharSeqLongSet::printSet()
//
//  PURPOSE: prints the set to file. The output consists of 'numseq' output 
//           lines, with the n-th line containing the n-th sequence followed by 
//           '\n' (n=0,...,numseq-1).
//
//  ARGUMENT:
//   ---> fp: pointer to the start of the first output line.
//
//  COMPLEXITY: O(numseq)
//
//
void CharSeq::CharSeqLongSet::printSet(FILE* fp) {

  for(long int s=0; s<numseq; ++s) 
    fprintf(fp, "%s\n", set[s]);
}
