/******************************************************************************/
/*                                                                            */
/*                                    mkprf                                   */
/*                                                                            */
/*             Builds a profile from a multiple sequence alignment.           */
/*                                                                            */
/* ARGUMENT:                                                                  */
/* ---> ALIFIL: path to the file containing the multiple sequence alignment.  */
/*              The file consists of N lines, with one aligned sequence per   */
/*              line. Each aligned sequence starts at the start of the line   */
/*              and is immediately followed by a line terminator. All aligned */
/*              sequences, including any gaps, have the same length. The file */
/*              terminates after the last aligned sequence's line terminator. */
/*                                                                            */
/* OUTPUT:                                                                    */
/* - the profile printed on screen.                                           */
/*     The rows refer to the residue positions in the alignment, while the    */
/*   columns refer to the amino acid types; the last column refers to the gap.*/
/*     The value of the profile at position i for amino acid type A is        */
/*   F(i,A(0))*S(A(0),A) + ... + F(i,A(19))*S(A(19),A), where A(j) is the     */
/*   j-th amino acid type and F(i,X) is the frequency of amino acid type X at */
/*   position i; S(X,Y) is the substitution score for amino acid types X and  */
/*   Y as determined from the substitution matrix at SUBSTMTX (i=0,...,M-1,   */
/*   where M is the length of the alignment; j=0,...,19)                      */
/*     The value of the profile at position i for the gap is -1/[0.01+F(i,-)],*/
/*   where F(i,-) is the gap frequency at position i.                         */
/*                                                                            */
/* NOTES:                                                                     */
/* - the input multiple sequence alignment is not necessarily one obtained by */
/*   a multiple sequence alignment algorithm. What is required is that there  */
/*   should be one sequence per line.                                         */
/* - the amino acid frequencies at each position are computed according to    */
/*   the Henikoff sequence weighting (J. Mol. Biol. 1994. 243, 574-578).      */
/*                                                                            */
/* ASSUMPTIONS:                                                               */
/* - the path to the substitution matrix used to generate the profile,        */
/*   specified in SUBSTMTX, is correct.                                       */
/*                                                                            */
/*   Remove the embeded path 2018.11.1. by YangCao                            */
/******************************************************************************/

#include "Profile.h"
#include "CharSeqLongSet.h"
#include "SubstMtx.h"

int main(int argc, char** argv) {

  if(argc<3) 
  {
      printf("Usage: mkprf [msa.txt] [path/blosum-62.txt]\n");
      return 0;
  }
  // read alignment
  FILE* fp = fopen(argv[1], "r");
  if(!fp) {
    printf("cannot find %s\n", argv[1]);
    return 0;
  }
  CharSeq::CharSeqLongSet msa(fp); 
  fclose(fp);

  // read substitution matrix
  //const char SUBSTMTX[] = "!BINDIR!/INPUT/subs/blosum-62.txt";
  
  Protein::SubstMatrix sbmtx(argv[2]);

  // build profile
  Protein::Profile prf(&msa, &sbmtx);
  prf.printPrf();
}
