/******************************************************************************/
/*                                                                            */
/*                               Profile.cpp                                  */
/*                                                                            */
/******************************************************************************/

#include "Profile.h"
#include "CharSeqLongSet.h"
#include "SubstMtx.h"

//
//
// CONSTRUCTOR: Protein::Profile::Profile()
//
// ARGUMENTS:
// ---> msa: pointer to a multiple sequence alignment
// ---> sbmtx: pointer to a substitution matrix
//
// PURPOSE: builds a profile from the multiple sequence alignment and the
//          substitution matrix.
//
//
Protein::Profile::Profile(const CharSeq::CharSeqLongSet* msa, 
                          const SubstMatrix* sbmtx) :
                          prows(msa->getSeqLen()), pcols(AMINOS+1) {

  buildAminoOccurMtx(msa);
  
  seqw = new float[msa->getNumSeq()];
  setHenikWei(msa);

  buildAminoFreqMtx();

  buildPrf(sbmtx);
}


//
//
// DESTRUCTOR: Protein::Profile::~Profile()
//
//
Protein::Profile::~Profile() {
  
  // delete profile matrix
  for(int i=0; i<prows; i++)
    delete[] prf[i];
  delete[] prf;

  // delete amino acid frequency matrix
  for(int i=0; i<prows; i++)
    delete[] afreq[i];
  delete[] afreq;

  // delete sequence weights
  delete[] seqw;

  // delete amino acid occurrence matrix
  for(int i=0; i<prows; i++) {
    for(int j=0; j<pcols; j++)
      delete aocc[i][j];  
    delete[] aocc[i];
  }
  delete[] aocc;
}


//
//
// FUNCTION: Protein::Profile::buildAminoOccurMtx()
//
// ARGUMENTS:
// ---> msa: pointer to a multiple sequence alignment
//
// PURPOSE: builds the matrix of amino acid occurrence from the alignment.
//
// 
void Protein::Profile::buildAminoOccurMtx(const CharSeq::CharSeqLongSet* msa) {

  using std::vector;

  // allocate and zero out the matrix
  aocc = new vector<long int>** [prows];
  for(int i=0; i<prows; ++i) {
    aocc[i] = new vector<long int>* [pcols];
  }
  for(int i=0; i<prows; ++i)
    for(int j=0; j<pcols; ++j)
      aocc[i][j] = 0;

  // fill in the matrix
  const long int MROWS = msa->getNumSeq();
  const int MCOLS = msa->getSeqLen();
  for(long int i=0; i<MROWS; ++i)
    for(int j=0; j<MCOLS; ++j) {

      char item = msa->get(i,j);
      int ami = (item == '-') ? AMINOS : charToAmino(item);

      vector<long int>** sqls = &(aocc[j][ami]);
      if(*sqls)
        (*sqls)->push_back(i);
      else
        *sqls = new vector<long int>(1, i);
    }
}


//
//
// FUNCTION: Protein::Profile::printAOM()
//
// PURPOSE: prints vertically to screen the amino acid occurrence matrix.
//
// 
void Protein::Profile::printAOM() const {

  for(int i=0; i<prows; ++i)
    for(int j=0; j<pcols; ++j) {
      printf("--- (%d,%d) ---\n", i, j);
      std::vector<long int>* ls = aocc[i][j];
      if(ls) {
        long int sz = ls->size();
        for(long int k=0; k<sz; k++)
          printf("%ld\n", (*ls)[k]);
      }
    }
}


//
//
// FUNCTION: Protein::Profile::buildAminoFreqMtx()
//
// PURPOSE: builds the matrix of amino acid frequencies.
//
// NOTES:
// - see Profile.h for how the frequencies are computed.
//
void Protein::Profile::buildAminoFreqMtx() {

  afreq = new float* [prows];
  for(int i=0; i<prows; ++i) {
    afreq[i] = new float[pcols];
  }

  for(int i=0; i<prows; ++i)
    for(int j=0; j<pcols; ++j) {
      std::vector<long int>* seqls = aocc[i][j];
      if(seqls) {
        float tot = 0;
        long int sz = seqls->size();
        for(int k=0; k<sz; ++k)
          tot += seqw[(*seqls)[k]];
        afreq[i][j] = tot;
      }
      else 
        afreq[i][j] = 0;
    }
}


//
//
// FUNCTION: Protein::printAFM()
//
// PURPOSE: prints to screen the matrix of amino acid frequencies.
//          The rows refer to the residue positions, the columns refer to
//          the amino acid types.
//
void Protein::Profile::printAFM() const {

  printAminosLine(9); printf("\n");

  for(int i=0; i<prows; ++i) {
    printf("%5d", i);
    for(int j=0; j<pcols; ++j)
      printf("%9f", afreq[i][j]);
    printf("\n");
  }

  printAminosLine(9); printf("\n");
}


//
//
// FUNCTION: Protein::Profile::setHenikWei()
//
// ARGUMENTS:
// ---> msa: pointer to the multiple sequence alignment
//
// PURPOSE: sets the sequence weights using the Henikoff method 
//          (as described in J. Mol. Biol. 1994. 243, 574-578)
//
//
void Protein::Profile::setHenikWei(const CharSeq::CharSeqLongSet* msa) {

  // count num. amino acid types (including gaps) for each residue position
  int* numaa = new int[prows];
  for(int i=0; i<prows; ++i) {
    int count = 0;
    for(int j=0; j<pcols; ++j) {
      if(aocc[i][j])
        count++;
    }

    numaa[i] = count;
  }

  // compute non-normalized weights
  long int MROWS = msa->getNumSeq();
  int MCOLS = msa->getSeqLen();
  for(long int i=0; i<MROWS; ++i) {

    float wei = 0;
    for(int j=0; j<MCOLS; j++) {

      char item = msa->get(i,j);
      int ami = (item == '-') ? AMINOS : charToAmino(item);
      std::vector<long int>* seqls = aocc[j][ami];
      
      wei += 1.0 / (numaa[j] * seqls->size());   
    }  
    seqw[i] = wei;
  }
  
  // normalize weights
  float tot = 0;
  for(long int i=0; i<MROWS; ++i)
    tot += seqw[i];

  for(long int i=0; i<MROWS; ++i)
    seqw[i] /= tot;

  delete[] numaa;
}


//
//
// FUNCTION: Protein::buildPrf()
//
// PURPOSE: builds the profile (see Profile.h for its exact contents).
//
// ARGUMENT:
// ---> sbmtx: pointer to the substitution matrix used to compute the profile.
// 
//
void Protein::Profile::buildPrf(const SubstMatrix* sbmtx) {

  // aminoToChar() as a fast look-up table
  const char A2C[] = "GPDEKRHSTNQAMYWVILFC";

  prf = new float* [prows];
  for(int i=0; i<prows; ++i)
    prf[i] = new float[pcols];

  // build the submatrix referring to the 20 amino acid types
  for(int i=0; i<prows; ++i)
    for(int j=0; j<AMINOS; ++j) {
      char cj = A2C[j];
      float tot = 0;
      for(int k=0; k<AMINOS; ++k) {
        char ck = A2C[k];
        tot += afreq[i][k] * sbmtx->getScore(cj, ck);
      }
      prf[i][j] = tot;
    }

  // build the submatrix referring to the gap
  for(int i=0; i<prows; ++i)
    prf[i][AMINOS] = -1.0 / (0.01 + afreq[i][AMINOS]);
}


//
//
// FUNCTION: Protein::printPrf()
//
// PURPOSE: prints the profile to screen.
//          The rows refer to the residue positions, the columns refer to
//          the amino acid types.
//
void Protein::Profile::printPrf() const {

  printAminosLine(12); printf("\n");

  for(int i=0; i<prows; ++i) {
    printf("%5d", i);
    for(int j=0; j<pcols; ++j)
      printf("%12f", prf[i][j]);
    printf("\n");
  }

  printAminosLine(12); printf("\n");
}


//
//
// FUNCTION: Protein::printAminosLine()
//
// PURPOSE: prints all amino-acid-type names on a single line of the screen.
//          The i-th name equals aminoToChar(i) for i=0,...,19, and '-' 
//          for i=20.
//
// ARGUMENT:
// ---> awid: width of the field for each amino acid type (in characters)
//
//
void Protein::printAminosLine(const int awid) {

  char format[5];
  sprintf(format, "%%%dc", awid);

  printf("%5c", ' ');
  for(int i=0; i<AMINOS; ++i)
    printf(format, aminoToChar(AminoName(i)));
  printf(format, '-');
}
