/******************************************************************************/
/*                                                                            */
/*                                 Profile.h                                  */
/*                                                                            */
/* A class to build a profile from a multiple sequence alignment and a        */
/* substitution matrix.                                                       */                           
/*                                                                            */
/******************************************************************************/

#ifndef PROFILE_H
#define PROFILE_H

#include <vector>
#include "AminoName.h"

namespace CharSeq {
  class CharSeqLongSet;
}

namespace Protein {

  class SubstMatrix;

  class Profile {

    int prows; // number of rows in the profile (residue positions)
    int pcols; // number of columns in the profile (amino acid types)

    // matrix of amino acid occurrence within the sequences.
    // aocc[i][j] lists the indices of the sequences at whose position i amino
    // acid type A(j) occurs (j=0,...,20, with A(j)=aminoToChar(j) for 
    // j=0,...,19 and A(j)='-' for j=20; i=0,...,N-1, where N
    // is the number of residue positions in the multiple sequence alignment).
    // For each list of sequence indices, the k-th index refers to the k-th 
    // sequence to meet the above condition within the multiple sequence 
    // alignment (in a scan of sequences from index 0 to index M-1, where M
    // is the number of sequences in the multiple sequence aligment, and from
    // residue 0 to residue N-1), for k=0,...,L-1, where L is the size of the 
    // list. aocc[i][j] = 0 if the list is empty.
    std::vector<long int>*** aocc;
    void buildAminoOccurMtx(const CharSeq::CharSeqLongSet* msa);
    void printAOM() const;

    // matrix of amino acid frequencies.
    // afreq[i][j] is the frequency of amino acid type A(j) at position i in 
    // the multiple sequence alignment. afreq[i][j] is computed as the sum of 
    // the weights of the sequences at whose position i amino acid type A(j) 
    // occurs (j=0,...,20, with A(j)=aminoToChar(j) for j=0,...,19 and A(j)='-' 
    // for j=20; i=0,...,N-1, where N is the number of residue positions in the
    // multiple sequence alignment). Note that the frequency matrix does not
    // depend on how the sequence weights are computed.
    float** afreq;
    void buildAminoFreqMtx();
    void printAFM() const;

    // the profile.
    // prf[i][j] is the value of the profile for amino acid type A(j) at residue 
    // position i of the multiple sequence alignment (j=0,...,20, with 
    // A(j)=aminoToChar(j) for j=0,...,19 and A(j)='-' for j=20; i=0,...,N-1, 
    // where N is the number of residue positions in the multiple sequence 
    // alignment). 
    // For j=0,...,19, prf[i][j] is computed as 
    // F(i,A(0))*S(A(0), A(j)) + ... + F(i,A(19))*S(A(19), A(j)), where 
    // F(i, A(k)) is the frequency of amino acid type A(k) at residue position i
    // and S(A(k), A(j)) is the substitution score for the (A(k), A(j)) pair of
    // amino acid types (k=0,...,19).
    // For j=20, prf[i][j] is -1/[0.01+F(i,-)], where F(i,-) is the gap 
    // frequency at position i.
    float** prf;
    void buildPrf(const SubstMatrix* sbmtx);

    // sequence weights: seqw[i] is the weight of sequence i, for i=0,...,M-1
    float* seqw; 
    void setHenikWei(const CharSeq::CharSeqLongSet* msa);

    public:
    Profile(const CharSeq::CharSeqLongSet* msa, const SubstMatrix* sbmtx);
    ~Profile();
    void printPrf() const;
  };

  void printAminosLine(const int awid);
}

#endif
