#include <stdlib.h>
#include "SubstMtx.h"
#include "AminoName.h"

//
//
// CONSTRUCTOR : Protein::SubstMatrix::SubstMatrix()
//
// PURPOSE : loads the substitution matrix from file.
//
// ARGUMENT:
//  ---> mtxf: pathname of the file which contains the matrix in
//             lower-triangular form. The zero's above the diagonal are not
//             represented. The first column contains the names of the amino
//             acid types in one-letter codes. The same applies to the last row.
//
// NOTES:
//  - this constructor can load any kind of matrix (e. g., BLOSUM, PAM, ...) as
//    long as the above format is respected.
//  - mtx[A(i)][A(j)] equals I[i][j], for i,j = 0,...,19. Here, I is the input 
//    substitution matrix, and A(k) is the integer representation of the amino
//    acid 1-letter code in position k of the input matrix.
//  
//
Protein::SubstMatrix::SubstMatrix(const char* mtxf) {

  FILE* fp = fopen(mtxf, "r");
  if(!fp) {
    printf("can't find %s\n", mtxf);
    exit(0);
  }

  AminoName aminoseq[20];

  char name[2];
  for(int i=0; i<20; ++i) {
    fscanf(fp, "%s", name);
    aminoseq[i] = charToAmino(name[0]);
    for(int j=0; j<=i; ++j) {
      AminoName row = aminoseq[i];
      AminoName col = aminoseq[j];
      fscanf(fp, "%lf", &mtx[row][col]);  
      mtx[col][row] = mtx[row][col];
    }
  }

  fclose(fp);
}


//
//
// FUNCTION: Protein::SubstMatrix::getScore()
//
// PURPOSE: returns the substitution score stored in the matrix for two amino
//          acid types.
//
// ARGUMENTS:
//  ---> a1: 1-letter code of the first amino acid type
//  ---> a2: 1-letter code of the second amino acid type
//
//
double Protein::SubstMatrix::getScore(const char a1, const char a2 ) const {

  return mtx[charToAmino(a1)][charToAmino(a2)];
}

