#!/usr/bin/env python
"""
The jobsubmit.py modules provides methods for submitting and running
jobs that use qsub and PBS directives.

JOBS = jobsubmit.JOBS()  #initializes module.
JOBS.SubmitJob(
    jobInput =['/pathToprogram/program','argument1','argument2',...],
    jobOutput=['/fullPath/ToOutputFile1','/fullPath/ToOutputFile2',..], 
    jobName="jobName", 
    outputDir=".")

    # jobInput is the program you want to run and its command line arguments.

    # jobOutput is a list of all the output files.  If all the files are completed
    # the jobsubmission is skipped.  If list is empty job will be submitted.

    # jobName is the job name which should be unique.  The module will check to
    # see if the program is currently submitted or running and if so will skip 
    # submission. By default it is name of your submitted program +".pbs".

    # outputDir is where the pbs script the standard error and standard output 
    # files from the jobsubmission are written to. By default it is the folder
    # of your subitted program
"""
# 2015-09-28 Chengxin forked from Brandon

#python 2.x and 3.x cross compatibility
from __future__ import print_function, division
import sys
if sys.version_info[0] >= 3:
    xrange = range

import os
import time
from subprocess import Popen, PIPE, call
import sys,re

def IsUserOverJobLimit(username='',maxJobs=300):
    """
    Checks if user over maximum number of jobs
    Arguments:
        username (str): name of current user
        maxJobs (int): Maximum number of jobs allowed to be submitted
    Returns:
        True if over maximum limit. False otherwise.
    """
    if not username:
        username=os.getenv("USER")
    maxJobs = int(maxJobs)
    p = Popen("qstat |grep "+username + " |wc -l",shell = True,
                stdout=PIPE,stderr=PIPE)
    stdout, stderr = p.communicate()
    if int(stdout) > maxJobs:
        return int(stdout)
    else:
        return False


class JOBS:
    """
    This class is for writing and sending jobs
    Attributes:
        username (str): username of the user submitting jobs
        walltime (str): hrs:min:sec Maximum length for job to run
        priority (str): Submission priority. choices are: 
                        default, urgent, casp, local, none
        maxJobs (int): Maximum number of jobs allowed for one user.
        AllMaxJobs (int): Maximum number of jobs allowed by all users.
    Methods:
        SubmitJob: Submits a Job.
        CreateJob: Creates a shell script for a job
        JobIsDoneOrRunning: Checks if job is completed or running
        Execute: Runs an executable command locally. Prints stdout in realtime.
            Returns stdout and stderror
        mkdir: Attempts to make a given directory
        KillJob: Kill a particular job
        KillAll: Kill all jobs of a user
        CleanTmp: Clean all temporary directories of a user on all nodes
    """

    def __init__(self,username='',walltime="24:59:00", priority='default',
            maxJobs =300, AllMaxJobs=1000, mem="3000mb"):
        self.username = username if username else os.getenv("USER")
        self.walltime = walltime.lstrip("walltime=").split(',')[0]
        self.priority = priority
        self.maxJobs  = maxJobs
        self.AllMaxJobs = AllMaxJobs

        if "mem=" in walltime:
            mem = walltime.split("mem=")[1]
        self.mem = mem.lstrip("mem=") if mem else "3000mb"

    def SubmitJob(self,jobInput=[], jobOutput=[], jobName='', outputDir='',
        checkInf='', PBSerror = '', PBSoutput = '',ForceSubmit = False):
        """
        Creates and submits a job if the job is not: in que, running or
        completed.

        Arguments:
            jobInput ([str]): List of commands use for running the script.
                ie. jobInput = ['/path/exampleJob', 'arg1','arg2'].
                Each item in the list should contain the full path. 
            jobOutput ([str]): Full path name for all files and directories
                that should be returned after job is completed. Used
                to for checking if job has finished.
            jobName (str): Name of job to be submitted.
            outputDir (str): Full path of output directory
            PBSerror (str,optional): Full path name of pbs error file. Default
                uses outputDir/jobName.err
            PBSoutput (str,optional): Full path name of pbs output file. Default
                uses outputDir/jobName.out
            ForceSubmit (bool, optional): Default is False.  Used to allow
                user to force a job submission if the output files 
                are present.  The output may be present but not complete
                if user detects this can force resubmission.
        """
        if isinstance(jobInput,basestring):
            jobInput=[jobInput]
        if isinstance(jobOutput,basestring):
            jobOutput=[jobOutput]
        if not jobName:
            jobName=os.path.basename(jobInput[0])+".pbs"
        if self.JobIsRunning(jobName):
            return jobName+" is running or in queue skip submission"

        if not outputDir:
            outputDir=os.path.dirname(os.path.abspath(jobInput[0]))
        else:
            outputDir=os.path.abspath(outputDir)
        if not checkInf:
            checkInf = self.username

        if not self.JobIsDone(jobName, jobOutput) or ForceSubmit:
            self.CreateJob(jobInput, outputDir, jobName, PBSerror = '',
                PBSoutput = '')
            
            if self.priority=="none": # do not run job
                print(jobName,"not run because priority=='none'")
                return jobName+" skipped"
            elif self.priority=="local": # run at head node
                p = Popen(outputDir + '/' +jobName, shell=True,
                    stdout=PIPE, stderr=PIPE)
            else:    
            #check number of jobs running, if over max wait for some jobs to finish
                period_printed=False
                while True:
                    #p = Popen("qstat |grep "+self.username + " |wc -l",
                    p = Popen("qstat -f |grep Job_Name |grep "+checkInf+ " |wc -l",
                        shell = True, stdout=PIPE,stderr=PIPE)
                    current_user_job_num, stderr = p.communicate()

                    p = Popen("qstat |wc -l",
                        shell = True, stdout=PIPE,stderr=PIPE)
                    all_job_num, stderr = p.communicate()

                    #if int(current_user_job_num) > self.maxJobs and int(all_job_num) > self.AllMaxJobs :
                    if int(current_user_job_num) > self.maxJobs: # and int(all_job_num) > self.AllMaxJobs :

                        if period_printed==False: # first period to print
                            period_printed=True
                            sys.stderr.write(str(current_user_job_num).strip() \
                                +" jobs submitted by "+self.username+ \
                                ". Waiting for jobs to complete.")
                        sys.stderr.write('.')
                        time.sleep(60)
                    else:
                        if period_printed: # line break if period was printed
                            sys.stderr.write('\n')
                        break    
                #run job
                p = Popen('qsub '+outputDir + '/' +jobName, shell=True,
                    stdout=PIPE, stderr=PIPE)

            stdout, stderr = p.communicate()
            out = stdout.split('\n')
            for line in out:
                if line != '':
                    print(line.strip())
            print(jobName, "submitted")
            time.sleep(0.1)
            return stdout.strip()
        else:
            return jobName+" completed skipping submission"

    def CreateJob(self,jobInput, outputDir, jobName, PBSerror = '',
                    PBSoutput = ''):
        """
        Creates a shell PBS script for job submission.

        Arguments:
            jobInput ([str]): A list containing the inputs for the job
                script. ie jobInput = ['./exampleJob.py','arg1','arg2',etc.]
            outputDir (str): The full path for the output directory.  Unless
                otherwise specified the job script, output and error files
                are written there.
            jobName (str): The name of the job without file extensions. ie
                'exampleJob' is okay but 'exampleJob.txt' not okay.
            PBSerror (str,optional) Full path name of PBS error file.  Default
                is the jobName.err stored in the outputDir
            PBSoutput (str,optional): Full path name of PBS output file. Default
                is the jobName.out stored in the outputDir 
        Returns:
            directory (directory): Attempts to create the output directory.
            PBS script (file): PBS job submission script. Which is written to
                outputDir/jobName   
        """
        #job inputs and outputDir require full paths
        if PBSerror == '':
            if jobName.endswith(".pbs"):
                PBSerror = outputDir + '/' + jobName[:-len(".pbs")] + ".err"
            else:
                PBSerror = outputDir + '/' + jobName + ".err"
        if PBSoutput =='':
            if jobName.endswith(".pbs"):
                PBSoutput = outputDir + '/' + jobName[:-len(".pbs")] + ".out"
            else:
                PBSoutput = outputDir + '/' + jobName + ".out"
        self.mkdir(outputDir)
        fjob = open(outputDir + '/' + jobName, 'w')
        fjob.write("#PBS -q "+self.priority +'\n')
        fjob.write("#PBS -d /tmp/"+'\n')
        fjob.write("#PBS -l nodes=1:ppn=1,walltime="+self.walltime+ \
            ",mem="+self.mem+'\n')
        fjob.write("#PBS -o "+PBSoutput +'\n')
        fjob.write("#PBS -e "+PBSerror + '\n')
        fjob.write("#PBS -N "+jobName +'\n')    
        for input in jobInput:
            fjob.write(str(input))
            fjob.write(' ')
        fjob.write('\n')
        fjob.close()
        try:
            os.chmod(os.path.join(outputDir, jobName),
                 os.stat(os.path.join(outputDir, jobName)).st_mode|0111)
        except Exception,error:
            sys.stderr.write(str(error))

    def JobIsDone(self,jobName, jobOutput = None):
        """
        Checks if the job has been completed by checking for output files.

        Arguments:
            jobName (str): Name of job.
            jobOutput ([str]): List of output files and directories from job.  
                The program just checks for existance.  It does not check
                if the files have actually completed. If list is empty
                function returns False.
        Returns:
            True: If all output files are generated
            False: otherwise
        """ 
        if jobOutput is None:
            return False
            
        if not jobOutput:
            return False

        for output in jobOutput:
            if not os.path.exists(output):
                return False
            if os.path.isfile(output) and not os.path.getsize(output):
                return False # enforce non zero output
        print(jobName, "completed skipping submission")
        return True


    def JobIsRunning(self,jobName,verbose=True):
        """
        Checks if the job is currently running.
        Arguments:
            jobName (str): Name of job.
            verbose (bool): If true prints statements
        Returns:
            True: If job is running or in que.
            False: otherwise
        """ 
        jobName=jobName.split(os.sep)[-1]
        p = Popen("qstat -f|grep Job_Name", shell = True, stdout = PIPE,
                    stderr = PIPE) 
        stdout, stderr = p.communicate()
        out = stdout.split('\n')
        for name in out:
            name=name.strip()[len("Job_Name ="):].strip()
            if jobName==name:
                if verbose:
                    print(jobName, "is running or in queue skip submission")
                return True
        return False


    def Execute(self, command, raiseStdError = True):
        """
        Runs a system command.  Prints stdout in real time. Returns stdout and 
        stderr.
        
        Arguments:
            command (str or [str]): System command can be run.  Can be a string 
                or a list.  ie command = 'ls -la' or command = ['ls','-la']
            raiseStdError (bool): Default True causes an error to be raised if 
                the executable writes output to standard error. If set to false.
                stderr is ignored.
        Returns:
            stdout (str): Standard output written by executable.
            stderr (str): Standard error written by program. 
        """
        if isinstance(command, list):
            cmd = ' '.join(command)
        else:
            cmd = command   
        p = Popen(cmd, shell=True, stdout=PIPE, stderr = PIPE)
        stdoutTmp = []
        for line in iter(p.stdout.readline, ''):
            print(line,end='')
            stdoutTmp.append(line)
        stderTmp = []
        for line in p.stderr:
            stderTmp.append(line)
        stdout = ''.join(stdoutTmp)
        stderr = ''.join(stderTmp)
        if raiseStdError and stderr != '':
            print(command, "Prints error messages")
            print(stderr)
            raise RuntimeError
        elif stderr != '':
            print(stderr)

        return stdout, stderr 


 

    def mkdir(self, directory):
        try:
            os.mkdir(directory)
        except:
            pass
    def KillJob(self,jobName='',jobID=''):
        """
        Kills the given job.

        Requires:
            jobName or JobID to be given
        Arguments:
            jobName (str,optional): Name of job to be killed.
            jobID   (str or integer,optional): Job ID of job to be killed  
        """
        if jobID != '':
            print(jobID, "killed")
            #call(['qdel','-p',str(jobID)])
            call(['qdel',str(jobID)])
        elif jobName != '':
            print(jobName, "killed")
            call(['qdel','-p',jobName])
        else:
            pass    
        
    def KillAll(self):
        """
        Kill all jobs for a particular user.
        
        Arguments:
            username (str): The user name of the user. 
        """
        p=Popen("qstat -u "+self.username, shell = True,stdout=PIPE,stderr=PIPE)
        stdout, stderr = p.communicate()
        out = stdout.split('\n')
        joblist = []
        for line in out:
            if line == '':
                continue
            if line[0].isdigit():
                jobinfo = line.split()
                #print (jobinfo)
                jobID = jobinfo[0].split('.')[0]
                jobName = jobinfo[3]
                #print(jobID,jobName)
                joblist.append([jobID,jobName])
               	#sys.exit() 
        for job in joblist:
            print('*******************',job[0])
            self.KillJob(jobID = job[0] )
    def CleanTmp(self):
        """
        Delete all /tmp/username directories and its subdirectories
        on each node.

        Arguments:
            username (str): The user name for the user.
        """
        call(["tentakel", "/bin/rm", "-rf", "/tmp/"+self.username])

###########################
def qnodes(option='all'):
    '''Return a list of PBS cluster nodes satisfying `option`

    option:
        all           - all nodes
        reachable     - nodes reachable by ping
        unreachable   - nodes unreachable by ping
        state         - return state of nodes instead of node names

        job-exclusive - the node is completely filled with jobs
        offline       - administratively down
        down          - powered off
        free          - not all processors are in use
    '''
    p = Popen("qnodes", shell=True,stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate()

    pattern=re.compile("^\w+\n\s*state\s*=\s*[-\w,]+?\n|"+ \
                    "\n\n\w+\n\s*state\s*=\s*[-\w,]+?\n")
    zlist=[re.split("\s*state\s*=\s*",e.strip()) \
        for e in pattern.findall(stdout)]
    node_list,state_list=map(list,zip(*zlist))

    if option=="state":
        return state_list

    elif option=="reachable": # do not try to ping every node
        zlist=[(n,s) for n,s in zlist if not "down" in s and not "offline" in s]
    elif option=="unreachable":
        zlist=[(n,s) for n,s in zlist if "down" in s or "offline" in s]  
    elif option=="job-exclusive": # job-exclusive
        zlist=[(n,s) for n,s in zip(node_list,state_list) if "job-exclusive" in s]
    elif option=="offline" in option: # offline
        zlist=[(n,s) for n,s in zip(node_list,state_list) if "offline" in s]
    elif option=="down": # down
        zlist=[(n,s) for n,s in zip(node_list,state_list) if "down" in s]
    elif option=="free": # free
        zlist=[(n,s) for n,s in zip(node_list,state_list) if "free" in s]

    if not zlist:
        node_list=[]
        state_list=[]
    else:
        node_list,state_list=map(list,zip(*zlist))
    return node_list


if __name__ == "__main__":
    if len(sys.argv)<2:
        docstring='''Querying PBS cluster nodes information

    option:
        all           - all nodes
        reachable     - nodes reachable by ping
        unreachable   - nodes unreachable by ping
        state         - return state of nodes instead of node names

        job-exclusive - the node is completely filled with jobs
        offline       - administratively down
        down          - powered off
        free          - not all processors are in use
    '''
        print(docstring)
        exit()
    print('\n'.join(qnodes(sys.argv[1])))
