package math;
import java.io.IOException;
import java.util.*;

import model.PDB;
import model.ProteinChain;
import model.Residue;
import model.Tern;

public class InertiaTensor {
	
	private ArrayList<Tern> ternList;
	private ArrayList<Float> massList;
	private float mass;
	private int pointNum;
	
	public InertiaTensor(ArrayList<Tern> ternList)
	{
		this.ternList = ternList;
		this.massList = new ArrayList<Float>();
		for(int i=0;i<ternList.size();i++) this.massList.add(1.0f);
		this.pointNum = this.ternList.size();
		this.mass = 1.0f*this.pointNum;
//		System.out.println("m=" +mass);
	}

	public InertiaTensor(ProteinChain pc)
	{
		ArrayList<Tern> crdList = new ArrayList<Tern>();
		for(Residue res : pc.getResidue())
		{
			if(res.getAtoms().containsKey("CA"))
				crdList.add(res.getAtom("CA").getCoordinate());
		}
		this.ternList = crdList;
		this.massList = new ArrayList<Float>();
		for(int i=0;i<ternList.size();i++) this.massList.add(1.0f);
		this.pointNum = this.ternList.size();
		this.mass = 1.0f*this.pointNum;
	}
	
	public double[] calInertiaTensor()
	{
		/*
		 * Inertia tensor:
		 *  a   -d   -e
		 * -d    b   -f
		 * -e   -f    c
		 * return {a,b,c,d,e,f}
		 */
		
		Tern cog = new Tern(0,0,0);
		for(int i=0;i<this.ternList.size();i++)
		{
			Tern t = this.ternList.get(i);
			cog = cog.add(t.multiply(this.massList.get(i)));
		}
		cog = cog.divide(this.mass);
		ArrayList<Tern> list2 = new ArrayList<Tern>();
		for(Tern t : this.ternList)
		{
			list2.add(t.subtract(cog));
		}
		double[] it = new double[6];
		for(int i=0;i<this.ternList.size();i++)
		{
			Tern t = list2.get(i);
			double m = this.massList.get(i);
			it[0] += (t.y()*t.y() + t.z()*t.z())*m;
			it[1] += (t.z()*t.z() + t.x()*t.x())*m;
			it[2] += (t.x()*t.x() + t.y()*t.y())*m;
			it[3] += t.x()*t.y()*m;
			it[4] += t.x()*t.z()*m;
			it[5] += t.y()*t.z()*m;
		}
		
		return it;
	}
	
	public double[] principalMomentOfInertia()
	{
		double[] it = calInertiaTensor();
		double a = it[0];
		double b = it[1];
		double c = it[2];
		double d = it[3];
		double e = it[4];
		double f = it[5];
		
		double A = -1;
		double B = a + b + c;
		double C = d*d + e*e + f*f - a*b - a*c - b*c;
		double D = a*b*c - 2*d*e*f - a*f*f - b*e*e - c*d*d;
		
		CubicEquation ce = new CubicEquation(A,B,C,D);
		double[] solve = ce.solve();
		if(solve == null) return null;
		else return solve;
	}
	
	public double[] radiusOfGyrations()
	{
		double[] pmi = principalMomentOfInertia();
		double[] rog = new double[3];
		rog[0] = Math.sqrt(pmi[0]/this.mass);
		rog[1] = Math.sqrt(pmi[1]/this.mass);
		rog[2] = Math.sqrt(pmi[2]/this.mass);
		return rog;
	}
	
	public double radiusOfGyration()
	{
		Tern cog = new Tern(0,0,0);
		for(int i=0;i<this.ternList.size();i++)
		{
			Tern t = this.ternList.get(i);
			cog = cog.add(t.multiply(this.massList.get(i)));
		}
		cog = cog.divide(this.mass);
		ArrayList<Tern> list2 = new ArrayList<Tern>();
		for(Tern t : this.ternList)
		{
			list2.add(t.subtract(cog));
		}
		double rr = 0;
		for(int i=0;i<this.ternList.size();i++)
		{
			Tern t = list2.get(i);
			double m = this.massList.get(i);
			rr += (t.x()*t.x() + t.y()*t.y() + t.z()*t.z())*m;		
		}
		rr = rr/this.mass;
		return Math.sqrt(rr);
	}
	
	public double shapeValue()
	{
		double[] rogs = radiusOfGyrations();
		double rr = this.radiusOfGyration()*this.radiusOfGyration();
		double rx = Math.sqrt(rr-rogs[2]*rogs[2]);
		double ry = Math.sqrt(rr-rogs[1]*rogs[1]);
		double rz = Math.sqrt(rr-rogs[0]-rogs[0]);
		return rz/rx;
	}
	
	public static void main(String[] args) throws IOException
	{
		PDB pdb = new PDB(args[0]);
		InertiaTensor it = new InertiaTensor(pdb.getChain('A'));
		double[] rogs  = it.radiusOfGyrations();
		double rog = it.radiusOfGyration();
		System.out.println(rogs[0] +" "+ rogs[1] + " " + rogs[2] + " " + rog);
		double flat = rogs[2]/rogs[0];
		System.out.println(flat);
		
		Tern A = new Tern(0,0,0);
		Tern B = new Tern(2,0,0);
		Tern C = new Tern(0,8,0);
		Tern D = new Tern(2,8,0);
		Tern E = new Tern(0,0,4);
		Tern F = new Tern(2,0,4);
		Tern G = new Tern(0,8,4);
		Tern H = new Tern(2,8,4);
		ArrayList<Tern> list = new ArrayList<Tern>();
		list.add(A);
		list.add(B);
		list.add(C);
		list.add(D);
		list.add(E);
		list.add(F);
		list.add(G);
		list.add(H);
		it = new InertiaTensor(list);
		
		rogs  = it.radiusOfGyrations();
		rog = it.radiusOfGyration();
	//	System.out.println(rogs[0] +" "+ rogs[1] + " " + rogs[2] + " " + rog);
		
	}
}
