package math;


public final class Jacobi
{
    private Jacobi() {}
    
    private static void rotate(double[][] vecA, int i, int j, int k, int l,
            double[] gh, double s, double tau)
    {
        //gh[0] is g in EGAD_Lib, gh[1] is h in EGAD_Lib
        gh[0] = vecA[i][j];
        gh[1] = vecA[k][l];
        
        vecA[i][j] = gh[0] - s*(gh[1] + gh[0]*tau);
        vecA[k][l] = gh[1] + s*(gh[0] - gh[1]*tau);
    }

    public static void compute(
    double[][] vecA,  // input matrix, destroyed by algorithm
    double[] vecD,   // returned eigenvalues
    double[][] vecV,  // returned eigenvectors
    int n){             // returned number of rotations required

    //  Computes all eigenvalues and eigenvectors of vecA real symmetric matrix
    //  vecA[0..n-1][0..n-1]. On output, elements of vecA above the diagonal are
    //  destroyed. vecD[0..n-1] returns the eigenvalues of vecA. vecV[0..n-1][0..n-1]
    //  is vecA matrix whose columns contain, on output, the normalized
    //  eigenvectors of vecA. iNumRot returns the number of Jacobi rotations
    //  that were required.  Note that vecD and vecV must be allocated prior
    //  to calling this function
    
    int max_sweeps = 50;

    // Declare iterators
    int j=0,         // iterator in inner loop for applying rotations
    iQ=0,             // column/q iterator
    iP=0,             // row/p iterator
    i=0;              // iterator over sweeps

    double threshold=0,    // Threshold for performing vecA pq rotation
    theta=0,          //rotation angle
    tau=0,            // sin(theta)/(1 + cos(theta))
    t=0,              // 1/(2*theta)
    sum=0,            // sum of off-diagonal elements
    s=0,              // sin(theta)
    h=0,              // a_pp - a_qq, which is used to determine theta
    g=0,              // 100*current off-diagonal element (used to check threshold)
    c=0;              // cos(theta)
    
    double[] gh = new double[2];//zzz, for rotate(...)

    double[] b=new double[n];
    double[] z=new double[n]; // This array will accumulate terms
                   // of the form tapq as in equation (11.1.14)

    // Initialize eigenvectors to identity matrix
    for (iP = 0; iP < n; ++iP) 
    { 
        for (iQ = 0; iQ < n; ++iQ)
            vecV[iP][iQ]=0.0;
        vecV[iP][iP]=1.0;
    }
    
    for (iP = 0; iP < n; ++iP) 
    {
        b[iP]=vecD[iP]=vecA[iP][iP];  // Initialize b and vecD to the diagonal of vecA  
        z[iP] = 0.0;                  // initialize z to 0
    }

    // Start counting rotations
    int iNumRot = 0;
    for (i = 0; i < max_sweeps; ++i) // loop over sweeps
    {
        // Sum off-diagonal elements (of one triangle)
        sum = 0.0;
        for (iP=0;iP<n-1;++iP) 
        { 
            for (iQ = iP+1; iQ< n; ++iQ)
                sum += Math.abs(vecA[iP][iQ]);
        }

        // Normal return: when off-diagonal elements sum to zero,
        //                within machine precision, we consider
        //                the matrix diagonalized.  For non-degenerate
        //                eigenvalues, we should have quadratic convergence
        //                to this state.

        if (sum == 0.0)
            return ; // success
        if (i < 3)
            threshold = 0.2*sum/(n*n);   //...on the 1st three sweeps.
        else
            threshold = 0.0;                //...thereafter.

        // Begin sweep over all possible pq rotations
        for (iP = 0; iP < n-1; ++iP)
        {
            for (iQ = iP+1; iQ < n; ++iQ) 
            {
                g = 100.0 * Math.abs(vecA[iP][iQ]);
            
                // After four sweeps, skip the rotation if 
                // the off-diagonal element is small.
                if((i > 3) &&
                    (Math.abs(vecD[iP])+g == Math.abs(vecD[iP])) &&
                    (Math.abs(vecD[iQ])+g == Math.abs(vecD[iQ])))
                {
                    vecA[iP][iQ] = 0.0;
                }
                // perform pq rotation
                else if (Math.abs(vecA[iP][iQ]) > threshold) 
                {
                    // check how big the rotation needs to be
                    h = vecD[iQ] - vecD[iP];

                    // Perform pre-calculations and apply rotation so as to
                    // minimize roundoff error.
                    if (Math.abs(h)+g == Math.abs(h))
                        t = (vecA[iP][iQ]) / h; 
                    else 
                    {
                        theta = 0.5*h / (vecA[iP][iQ]); // Equation (11.1.10).
                        t = 1.0 / (Math.abs(theta) + Math.sqrt(1.0 + theta*theta) );
                        if (theta < 0.0)
                            t = -t;
                    }
                    c = 1.0 / Math.sqrt(1.0 + t*t);  // cos(theta)
                    s = t * c;              // sin(theta)
                    tau = s / (1.0 + c);      
                    h = t * vecA[iP][iQ];
                    z[iP] -= h;
                    z[iQ] += h;
                    vecD[iP] -= h;
                    vecD[iQ] += h;
                    vecA[iP][iQ] = 0.0;
                    for (j = 0; j < iP; ++j) // Case of rotations 0 <= j <p.
                        rotate(vecA,j,iP,j,iQ,gh,s,tau);
                    for (j = iP+1; j < iQ; ++j) // Case of rotations p <j<q.
                        rotate(vecA,iP,j,j,iQ,gh,s,tau);
                    for (j = iQ+1; j < n; ++j) // Case of rotations q <j < n.
                        rotate(vecA,iP,j,iQ,j,gh,s,tau);
                    for (j = 0; j < n; ++j) // update eigenvectors
                        rotate(vecV,j,iP,j,iQ,gh,s,tau);
                    ++iNumRot; // count this rotation
                }
            }
        }

        // Tidy up after current sweep
        for (iP = 0; iP < n; ++iP) 
        {
            b[iP] += z[iP];
            vecD[iP] = b[iP];     // Update vecD with the sum of tapq,
            z[iP]= 0.0;     // and reinitialize z.
        }
    }
    throw new RuntimeException("Error in Jacobi: Too many iterations in routine Jacobi");
    }
}



