package math;


import java.util.Arrays;


/**
 * @author chzhou
 *
 */
public class Matrix
{
    private double[][] mtx;
    
    private Matrix() {}
    
    public Matrix(int m, int n)
    {
        mtx = new double[m][n];
    }
    
    public Matrix(double[][] values)
    {
        final int rows = values.length;
        final int columns = values[0].length;
        mtx = new double[rows][];
        for(int i=0; i<rows; i++)
        {
            if(values[i].length != columns)
                throw new IllegalArgumentException("All rows must have the same length");
            mtx[i] = values[i].clone();
        }
    }
    
    public Matrix(Matrix m)
    {
        this.mtx = getCopyOf(m.mtx);
    }
    
    public static double[][] getCopyOf(double[][] a)
    {
        double[][] rt = new double[a.length][];
        for(int i=0; i<a.length; i++)
            rt[i] = a[i].clone();
        return rt;
    }
    
    public static Matrix asMatrix(double[][] values)
    {
        Matrix rt = new Matrix();
        rt.mtx = values;
        return rt;
    }
    
    public static void copy(double[][] src, double[][] dest)
    {
        for(int i=0; i<src.length; i++)
            System.arraycopy(src[i], 0, dest[i], 0, src[i].length);
    }
    
    public void copyValuesFrom(Matrix b)
    {
        copy(b.mtx, this.mtx);
    }
    
    public double get(int i, int j)
    {
        return mtx[i][j];
    }
    
    public void set(int i, int j, double value)
    {
        mtx[i][j] = value;
    }
    
    public int rows() { return mtx.length; }
    
    public int columns() { return mtx[0].length; }
    
    public double[][] getArray() { return mtx; }
    
    public boolean isSquare()
    {
        return mtx.length == mtx[0].length;
    }
    
    /**
     * Create an identity matrix
     */
    public static Matrix identityMatrix(int n)
    {
        Matrix rt = new Matrix(n, n);
        for(int i=0; i<n; i++) rt.set(i, i, 1.0);
        return rt;
    }
    
    public Matrix transpose()
    {
        int m = this.rows();
        int n = this.columns();
        Matrix rt = new Matrix(n, m);
        double[][] source = this.mtx;
        double[][] dest = rt.mtx;
        for(int i=0; i<m; i++)
            for(int j=0; j<n; j++)
                dest[j][i] = source[i][j];
        return rt;
    }
    
    public Matrix add(Matrix b)
    {
        final int M = this.rows();
        final int N = this.columns();
        if(M != b.rows() || N != b.columns())
            new IllegalArgumentException("Matrix dimensions must agree");
        
        Matrix rt = new Matrix(M,N);
        for(int i=0; i<M; i++)
            for(int j=0; j<N; j++)
                rt.mtx[i][j] = this.mtx[i][j] + b.mtx[i][j];
        return rt;
    }
    
    public Matrix subtract(Matrix b)
    {
        final int M = this.rows();
        final int N = this.columns();
        if(M != b.rows() || N != b.columns())
            new IllegalArgumentException("Matrix dimensions must agree");
        
        Matrix rt = new Matrix(M, N);
        for(int i=0; i<M; i++)
            for(int j=0; j<N; j++)
                rt.mtx[i][j] = this.mtx[i][j] - b.mtx[i][j];
        return rt;
    }
    
    public Matrix negate()
    {
        final int M = this.rows();
        final int N = this.columns();
        Matrix rt = new Matrix(M, N);
        for(int i=0; i<M; i++)
            for(int j=0; j<N; j++)
                rt.mtx[i][j] = -this.mtx[i][j];
        return rt;
    }
    
    public Matrix multiply(double d)
    {
        final int M = this.rows();
        final int N = this.columns();
        Matrix rt = new Matrix(M, N);
        for(int i=0; i<M; i++)
            for(int j=0; j<N; j++)
                rt.mtx[i][j] = d * this.mtx[i][j];
        return rt;
    }
    
    /**
     * See Jama (a java matrix package).
     */
    public Matrix multiply(Matrix b)
    {
        Matrix rt = new Matrix(this.rows(), b.columns());
        multiply(this.mtx, b.mtx, rt.mtx);
        return rt;
    }
    
    static void multiply(double[][] a, double[][] b, double[][] result)
    {
        final int Z = a[0].length;
        if(Z != b.length)
            throw new IllegalArgumentException("Matrix inner dimensions must agree");
        final int M = a.length;
        final int N = b[0].length;

        double[][] c = result;
        double[] bColumnJ = new double[Z];

        for (int j = 0; j < N; j++) 
        {
            for (int k = 0; k < Z; k++) 
            {
                bColumnJ[k] = b[k][j];
            }
            for (int i = 0; i < M; i++) 
            {
                double[] aRowI = a[i];
                double s = 0;
                for (int k = 0; k < Z; k++) 
                {
                    s += aRowI[k]*bColumnJ[k];
                }
                c[i][j] = s;
            }
        }
    }
    
    /**
     * Matrix inverse.
     * http://www.personal.leeds.ac.uk/~bgy1mm/Bioinformatics/rmsd.html
     */
    public void inverse()
    {
        final int N = mtx.length;
        boolean[] ipiv = new boolean[N];
        int[] indexColumn = new int[N];
        int[] indexRow = new int[N];

        for (int i = 0; i < N; i++)
        {
            double big = 0.0;
            int irow = 0, icol = 0;
            //find biggest element 
            for (int j = 0; j < N; j++) 
            {
                if (ipiv[j]) continue;
                for (int k = 0; k < N; k++)
                {
                    if (ipiv[k] || Math.abs(mtx[j][k])<big) continue;
                    big = Math.abs(mtx[j][k]);
                    irow = j;
                    icol = k;
                }
            }
            ipiv[icol] = true;

            if (irow != icol)
            {
                //swap row[irow] and row[icol]
                for (int l = 0; l < N; l++)
                {
                    double temp = mtx[irow][l];
                    mtx[irow][l] = mtx[icol][l];
                    mtx[icol][l] = temp;
                }
            }
            indexRow[i] = irow;
            indexColumn[i] = icol;

            //if biggest element is zero matrix is singular, bail
            if (mtx[icol][icol] == 0)
                throw new ArithmeticException("singular matrix, no inverse");

            double pinv = 1.0 / mtx[icol][icol];
            mtx[icol][icol] = 1.0;

            for (int l = 0; l < N; l++)
                mtx[icol][l] *= pinv;
            for (int ll = 0; ll < N; ll++)
            {
                if (ll == icol) continue;
                double dum = mtx[ll][icol];
                mtx[ll][icol] = 0.0;
                for (int l = 0; l < N; l++)
                    mtx[ll][l] -= mtx[icol][l] * dum;
            }
        }

        /* unscramble matrix */
        for (int l=N-1;l>=0;l--)
        {
            if (indexRow[l] != indexColumn[l])
            {
                for (int k=0;k<N;k++)
                {
                    double temp = mtx[k][indexRow[l]];
                    mtx[k][indexRow[l]] = mtx[k][indexColumn[l]];
                    mtx[k][indexColumn[l]] = temp;
                }
            }
        }
    }
    
    public void print()
    {
        for(double[] a : mtx)
        {
        	System.out.print("[");
        	for(double x : a)
        	{
        		System.out.printf("%5.2f ",x);
        	}
        	System.out.println("]");
        }
        System.out.println();    
    }
    
    
//code above this line is from Zhouchanghai
    
    public void exchangeTwoRow(int i, int j)
    {
    	if(i+1 > this.mtx.length || j+1 > this.mtx.length)
    		throw new RuntimeException("illegal operation");

    	for(int k=0;k<this.mtx[0].length;k++)
    	{
    		double tmp = this.mtx[i][k];
    		this.mtx[i][k] = this.mtx[j][k];
    		this.mtx[j][k] = tmp;
    	}
    }
    
    public void rowMultipleByConstant(int i, double lamda)
    {
    	if(i+1 > this.mtx.length || lamda == 0.0)
    		throw new RuntimeException("illegel operation");
    	
    	for(int k=0;k<this.mtx[0].length;k++)
    	{
    		this.mtx[i][k] = this.mtx[i][k]*lamda;
    	}
    }
    
    public void rowIAddByLamdaTimesRowJ(int i,int j,double lamda)
    {
    	if(i+1 > this.mtx.length || j+1 > this.mtx.length)
    		throw new RuntimeException("illegal operation");
    	
    	for(int k=0;k<this.mtx[0].length;k++)
    	{
    		this.mtx[i][k] += lamda*this.mtx[j][k];
    	}
    }
    
    public void gaussElimination()
    {
    	int rowNum = this.mtx.length;
    	int arrayNum = this.mtx[0].length;
    	
    	int min = rowNum;
    	if(rowNum > arrayNum)
    	{
    		min = arrayNum;
    	}
    	
    	for(int i=0;i<min;i++)
    	{
    		

   			int majorRow = i;
   			double maxValue = Math.abs(this.mtx[i][i]);
   			for(int j = i+1;j<min;j++)
   			{
   				if(Math.abs(this.mtx[j][i]) > maxValue)
   				{
   					maxValue = Math.abs(this.mtx[j][i]);
   					majorRow = j;
   				}
   			}
   			
   			if(maxValue > 0.0000000001)
   			{
   				if(i!=majorRow)
   				{
   					exchangeTwoRow(i,majorRow);
   				}
   		
   				rowMultipleByConstant(i,1.0/this.mtx[i][i]);
   				for(int j=i+1;j<min;j++)
   				{
   					if(this.mtx[j][i] != 0)
   					{
   						rowIAddByLamdaTimesRowJ(j,i,-1.0*this.mtx[j][i]);
    				}
    			}
    		}
    	}
    }
    
    
    
    public double[] solveLinearEquation()
    {
    	if(this.mtx.length+1 != this.mtx[0].length)
    	{
    		throw new RuntimeException("matrix size illegal " + (this.mtx.length) + " " + (this.mtx[0].length));
    	}
    	
    	this.gaussElimination();
    	int size = this.mtx.length;
    	for(int i=0;i<size;i++)
    	{
    		if(Math.abs(this.mtx[i][i]) < 0.0000000001)
    		{
//    			System.err.println("no unique solution");
//    			return null;
    			throw new RuntimeException("no unique solution");
    		}
    	}
    	
    	double[] sol = new double[size];
    	for(int i=size-1;i>=0;i--)
    	{
    		sol[i] = this.mtx[i][size];
    		for(int j=i+1;j<size;j++)
    		{
    			sol[i] -= this.mtx[i][j]*sol[j];
    		}
    	}
    	return sol;
    }
    
    public static void main(String[] args)
    {
    	double[][] a = {{1,2,3},{1,4,2},{3,1,4}};
    	Matrix m = new Matrix(a);
    	m.print();
    	Matrix mi = new Matrix(m);
    	mi.inverse();
    	
    	Matrix m2 = m.multiply(mi);
    	m2.print();
    	
    	
    	m.inverse();
    	m.print();
    	m.gaussElimination();
    	m.print();
    }
}

