package math;
import java.util.ArrayList;
import java.util.List;
import model.Tern;
/**
 * Kabsch's algorithm to superpose to structures.
 * http://www.personal.leeds.ac.uk/~bgy1mm/Bioinformatics/rmsd.html 
 * Kabsch, Wolfgang, (1976) "A solution of the best rotation to 
 * relate two sets of vectors", Acta Crystallographica 32:922
 * http://en.wikipedia.org/wiki/Kabsch_algorithm
 * @author chzhou
 * 
 */
public class SuperPose
{
    public static Tern getCentroid(List<Tern> points)
    {
        double x = 0, y = 0, z = 0;
        for (Tern t : points)
        {
            x += t.x();
            y += t.y();
            z += t.z();
        }
        int n = points.size();
        return new Tern(x / n, y / n, z / n);
    }
    
    /**
     * Calculate rmsd between two structures.
     * @param v1 first set of points, keep fixed
     * @param v2 second set of points
     * @param outMatrix transfrom matrix for output
     * @return rmsd score
     */
    public static double rmsd(List<Tern> v1, List<Tern> v2, double[][] outMatrix)
    {
        final int N = v1.size();
        Tern cent1 = getCentroid(v1);
        Tern cent2 = getCentroid(v2);
        List<Tern> temp1 = new ArrayList<Tern>(N);
        List<Tern> temp2 = new ArrayList<Tern>(N);
        for(int i=0;i<N;i++)
        {
            temp1.add(v1.get(i).subtract(cent1));
            temp2.add(v2.get(i).subtract(cent2));
        }
        Matrix tmtx = getAlignMatrix(temp1, temp2);

        Matrix move1 = getTransMatrix(cent2.negate());
        Matrix tempmtx = move1.multiply(tmtx);
        Matrix move2 = getTransMatrix(cent1);
        tmtx = tempmtx.multiply(move2);
        
        temp2.clear();
        for(Tern t : v2)
            temp2.add(multiply(tmtx, t));
        double answer = simpleRmsd(v1, temp2);
        Matrix.copy(tmtx.getArray(), outMatrix);
        return answer;
    }
    
    public static double simpleRmsd(List<Tern> v1, List<Tern> v2)
    {
        final int N = v1.size();
        if(N != v2.size())
            throw new IllegalArgumentException();
        double rt = 0;
        for(int i=0; i<N; i++)
            rt += v1.get(i).distSquare(v2.get(i));
        return Math.sqrt(rt/N);
    }
    
    private static Matrix getTransMatrix(Tern t)
    {
        Matrix rt = Matrix.identityMatrix(4);
        double[][] a = rt.getArray();
        a[3][0] = t.x();
        a[3][1] = t.y();
        a[3][2] = t.z();
        return rt;
    }
    
    /**
     * Multiply a point by a matrix.
     * @param matrix a 4x4 transform matrix
     * @param point the point to be transformed
     */
    private static Tern multiply(Matrix matrix, Tern point)
    {
        double[] ans = new double[4];
        for(int i=0;i<4;i++)
        {
          for(int j=0;j<3;j++)
          {
            ans[i] += point.get(j) * matrix.get(j, i);
          }
          ans[i] += matrix.get(3, i);
        }
        return new Tern(ans[0],ans[1],ans[2]);
    }
    
    /**
     * Calculate the matrix needed to align two structures. 
     * Notes: only calculates rotation part of matrix.
     * Assumes input has been aligned to centroids
     * @param v1 reference structure
     * @param v2 structure to align
     */
    public static Matrix getAlignMatrix(List<Tern> v1, List<Tern> v2)
    {
        final int N = v1.size();
        double correction = absMax(v1) * absMax(v2);

        Matrix A = new Matrix(4,4);
        A.set(3, 3, 1.0);
        double[][] array = A.getArray(); 
        for(int k=0;k<N;k++)
            for(int i=0;i<3;i++)
                for(int j=0;j<3;j++)
                    array[i][j] += (v1.get(k).get(i) * v2.get(k).get(j))/correction;

        Matrix At = A.transpose();
        Matrix Ainv = new Matrix(A);
        Ainv.inverse();
        Matrix temp = At.multiply(A);
        temp = sqrt(temp);
        return temp.multiply(Ainv);
    }

    /*
     * Check two matrices for near-enough equality
     */
    private static boolean almostEqual(Matrix a, Matrix b)
    {
        final int M = a.rows();
        final int N = a.columns();
        final double epsilon = 0.001;

        for (int i = 0; i < M; i++)
            for (int j = 0; j < N; j++)
                if (Math.abs(a.get(i, j) - b.get(i, j)) > epsilon)
                    return false;
        return true;
    }

    /**
     * Calculate the square root of a matrix using Denman-Beavers iteration.
     */
    public static Matrix sqrt(Matrix matrix)
    {
        final int N = matrix.rows();
        Matrix Y = new Matrix(matrix);
        Matrix Z = Matrix.identityMatrix(N);
        Matrix Y1 = new Matrix(N, N);
        Matrix Z1 = new Matrix(N, N);
        Matrix Y2 = new Matrix(N, N);
        Matrix invY = new Matrix(N, N);
        Matrix invZ = new Matrix(N, N);
        
        int iter = 0;
        do
        {
            invY.copyValuesFrom(Y);
            invZ.copyValuesFrom(Z);
            invY.inverse();
            invZ.inverse();

            for(int i=0;i<N;i++)
                for(int j=0;j<N;j++)
                {
                    Y1.set(i, j, 0.5*(Y.get(i,j) + invZ.get(i,j)));
                    Z1.set(i, j, 0.5*(Z.get(i,j) + invY.get(i,j)));
                }
            Matrix tmp = Y; 
            Y = Y1; Y1 = tmp;
            tmp = Z; Z = Z1; Z1 = tmp;
            Matrix.multiply(Y.getArray(), Y.getArray(), Y2.getArray());
        }
        while(!almostEqual(Y2, matrix) && iter++ < 20 );
        return Y;
    }
    
    private static double absMax(List<Tern> points)
    {
        double max = 0;
        for(Tern t : points)
        {
            double tmp = 0;
            if((tmp=Math.abs(t.x())) > max) max = tmp;
            if((tmp=Math.abs(t.y())) > max) max = tmp;
            if((tmp=Math.abs(t.y())) > max) max = tmp;
        }
        return max;
    }
    
}


