package math;

import java.io.Serializable;
import model.Tern;

/**
 A <code>Transform</code> is an object that stores the information 
 necessary to rotate and translate a set of points in 3D space.
 
 <p>A <code>Transform</code> holds a 3x4 transformation matrix used to 
 rotate and translate points in 3D space. The 3x4 matrix is composed 
 of a 3x3 rotation matrix and a 1x3 translation component.
 
 <p><code>Transform</code> can be constructed from a large number of 
 common representations for transformations, including specifying 
 matrix values directly, using a 1x4 quaternion or an angle-axis 
 rotation.
 
 <p>Since the angle-axis rotation requires the same parameters as the
 quaternion, building from an angle-axis pair must be performed
 through the BuildFromAngleAxis() function.
 */
public final class Transform implements Serializable
{
	private static final long serialVersionUID = -6964231163393364180L;
	
	private double r00, r01, r02, // 3x3 rotation matrix
                   r10, r11, r12,
                   r20, r21, r22,
                    tx,  ty,  tz; // 1x3 translation

    /**
     Constructs a <code>Transform</code> that preserves the 
     location of any transformed points.
     */
    public Transform() { r00=1; r11=1; r22=1; }

    /**
     Copy constructor.
     */
    public Transform(Transform t)
    {
        r00=t.r00; r01=t.r01; r02=t.r02;
        r10=t.r10; r11=t.r11; r12=t.r12;
        r20=t.r20; r21=t.r21; r22=t.r22;
        tx=t.tx; ty=t.ty; tz=t.tz;
    }

    /**
     Construct from 3x4 Transform matrix. r00-r09 is 3x3 rotation matrix. 
     (tx,ty,tz) is translation vector.
     */
    protected Transform(
            double r00,double r01,double r02,
            double r10,double r11,double r12,
            double r20,double r21,double r22,
            double  tx,double  ty,double  tz)
    {
        setRotation(r00,r01,r02, r10,r11,r12, r20,r21,r22);
        setTranslation(tx,ty,tz);
    }

    /**
     Build a <code>Transform</code> from a quaternion.
     <p>Quaternions are a common representation for transformation information
     in games and other applications that require fast angle-axis rotations
     or interpolation. This function will replace the transformation matrix
     with one that reproduces the behavior of this quaternion.
     
     <p>Implementation based on equation 10.23 on pg 187 of
     3D Math Primer for Graphics and Game Development, Dunn and Parberry.
     
     @param q vector component of the quaternion
     @param w scalar component of the quaternion
     */
    public static Transform buildFromQuaternion(Tern q,double w)
    {
        double X2 = 2.0 * q.x() * q.x();
        double Y2 = 2.0 * q.y() * q.y();
        double Z2 = 2.0 * q.z() * q.z();
        double WX = 2.0 * w * q.x();
        double WY = 2.0 * w * q.y();
        double WZ = 2.0 * w * q.z();
        double XY = 2.0 * q.x() * q.y();
        double XZ = 2.0 * q.x() * q.z();
        double YZ = 2.0 * q.y() * q.z();

        Transform t = new Transform();
        t.r00 = 1.0 - Y2 - Z2; // Diagonal members
        t.r11 = 1.0 - X2 - Z2; // "
        t.r22 = 1.0 - X2 - Y2;  // "
        t.r01 = XY + WZ; t.r02 = XZ - WY; // Off-diagonal members
        t.r10 = XY - WZ; t.r12 = YZ + WX; // "
        t.r20 = XZ + WY; t.r21 = YZ - WX; // "
        
        return t;
    }

    /**
     Replace the current <code>Transform</code> with an angle-axis rotation.
     
     <p>Rotating an object around an arbitrary axis is a common task in 
     many applications. The axis is specified by a vector from the origin 
     to the point <i>axis</i> (first parameter). The angle <i>theta</i> is, 
     as usual, in radians.
     */
    public static Transform buildFromAngleAxis(Tern axis,double theta)
    {
        double thetaOverTwo = theta/2.0;
        return buildFromQuaternion(
        		axis.multiply((float)Math.sin(thetaOverTwo)), 
                Math.cos(thetaOverTwo));
    }
    
    /**
     Set the rotation portion of the Transform Matrix.
     This leaves the translation portion of the matrix intact.
     */
    public void setRotation(
            double r00,double r01,double r02,
            double r10,double r11,double r12,
            double r20,double r21,double r22)
    {
        this.r00 = r00; this.r01 = r01; this.r02 = r02;
        this.r10 = r10; this.r11 = r11; this.r12 = r12;
        this.r20 = r20; this.r21 = r21; this.r22 = r22;
    }

    /**
     Retrieve the translation portion of the Transform matrix.
     */
    public Tern getTranslation()
    {
        return new Tern(tx,ty,tz);
    }

    /**
     Set the translation portion of the Transform matrix.
     This leaves the rotation portion of the matrix intact.
     */
    public void setTranslation(double tx,double ty,double tz)
    {
        this.tx = tx; this.ty = ty; this.tz = tz;
    }

    /**
     Set all matrix elements to zero.
     */
    protected void zeroMatrix()
    {
        zeroRotation();
        zeroTranslation();
    }

    /**
     Set all rotational matrix elements to zero, 
     leaving translation intact.
     */
    protected void zeroRotation()
    {
        r00 = r01 = r02 = r10 = r11 = r12 = r20 = r21 = r22 = 0;
    }

    /**
     Set all translational matrix elements to zero, 
     leaving rotation intact.
     */
    protected void zeroTranslation()
    {
        tx = ty = tz = 0;
    }
   
    /**
     Apply Transform to a <code>Tern</code> point.
     @param p the point to be transformed
     @return the new position after transform
     */
    public Tern transform(Tern p)
    {
        return new Tern(
                p.x()*r00 + p.y()*r01 + p.z()*r02 + tx,
                p.x()*r10 + p.y()*r11 + p.z()*r12 + ty,
                p.x()*r20 + p.y()*r21 + p.z()*r22 + tz);
    }

    /**
     Translate a point without rotation.
     @param p the point to be translated
     @return the new position
     */
    public Tern translate(Tern p)
    {
        return new Tern(p.x()+tx, p.y()+ty, p.z()+tz);
    }

    /**
     Build a Transform between two structures.

     <p>Uses least squares alignment to build a Transform that superimposes
     the second structure onto the first. Both structures are unmodified
     during the build process.
     
     <p>For more information about least squares alignment, see:
     Kearsley, Acta. Cryst. (1989) A45:208
     */
    public static Transform buildFromPoints(Tern[] newPos, Tern[] oldPos)
    {
    	if(newPos.length != oldPos.length)
    		throw new RuntimeException("must with same point number");
        
        if(newPos.length < 3) 
            throw new RuntimeException("requires at least 3 points");

    	// Step 1: Compute the centroid of each structure
        Tern newCenter = Tern.ZERO;
        Tern oldCenter = Tern.ZERO;

        for(Tern t : newPos) { newCenter = newCenter.add(t); }
        for(Tern t : oldPos) { oldCenter = oldCenter.add(t); }

        newCenter = newCenter.divide(newPos.length);
        oldCenter = oldCenter.divide(oldPos.length);

        // Step 2: Pass the centroids to the Build Function
        return buildFromPoints(newPos, oldPos, newCenter, oldCenter);
    }

    /**
     Build a Transform between two structures.
     
     <p>This form of Build() is similar to its counterpart, except it
     allows the specification of centroids for each structure. This
     can be useful when one wishes to rotate about a specific point
     (such as the C-alpha of an amino acid) when aligning the structures.
     
     <p>Otherwise, this function behaves identically to the other Build().
     */
    public static Transform buildFromPoints(
            Tern[] newPos, Tern[] oldPos, Tern newCenter, Tern oldCenter)
    {
    	if(newPos.length != oldPos.length)
    		throw new RuntimeException("must with same point number");
        
    	if(newPos.length < 3) 
            throw new RuntimeException("requires at least 3 points");
        
        // Step 1: Copy the points into a temporary vector for internal
        //         manipulation.
        final int size = newPos.length; 
        Tern[] tempNew = new Tern[size];
        Tern[] tempOld = new Tern[size];

        for(int i=0; i<size; i++)
        {
        	tempNew[i] = newPos[i].subtract(newCenter);
        	tempOld[i] = oldPos[i].subtract(oldCenter);
        }

        // Step 2: Build the rotation matrix
        Transform retVal = buildRotation(tempNew,tempOld);

        // Step 3: Figure out the correct translation. In this case, you want
        //  to first translate structure 2 to the origin, then rotate, then
        //  translate to the origin of structure 1. The math of the transform
        //  object is implemented as "rotate, then translate." We have to move
        //  the 1st translation so that it is performed after the rotation:
        Tern combTrans = Tern.ZERO.subtract(oldCenter);   // 1st translation
        //applying the rotation matrix to the translation is how we move the translation
        combTrans = retVal.transform(combTrans); 
        combTrans = combTrans.add(newCenter);   // combine it with the 2nd translation
        retVal.setTranslation(combTrans.x(), combTrans.y(), combTrans.z());
        return retVal;
    }

    /**
     Used by the build(..) functions to calculate the rotation matrix
     between two centered structures. Discards all previous rotation info.
     */
    public static Transform buildRotation(Tern[] points1, Tern[] points2)
    {
        Tern m,p;
        double[][] Q = new double[4][4]; //all zero
        double[][] R = new double[4][4]; //..
        double[] ev = new double[4];
        
       
        // Calculate the P matrix from each pair of points
        for(int i=0;i<points1.length && i<points2.length; i++)
        {

            m = points1[i].subtract(points2[i]);    // vector between the points
            p = points1[i].add(points2[i]);    // sum of position vectors
 
            Q[0][0] += m.x()*m.x()+m.y()*m.y()+m.z()*m.z();
            Q[1][1] += p.y()*p.y()+p.z()*p.z()+m.x()*m.x();
            Q[2][2] += p.x()*p.x()+p.z()*p.z()+m.y()*m.y();
            Q[3][3] += p.x()*p.x()+p.y()*p.y()+m.z()*m.z();
            Q[0][1] += p.y()*m.z()-m.y()*p.z(); Q[1][0] = Q[0][1];
            Q[0][2] += m.x()*p.z()-p.x()*m.z(); Q[2][0] = Q[0][2];
            Q[0][3] += p.x()*m.y()-m.x()*p.y(); Q[3][0] = Q[0][3];
            Q[1][2] += m.x()*m.y()-p.x()*p.y(); Q[2][1] = Q[1][2];
            Q[1][3] += m.x()*m.z()-p.x()*p.z(); Q[3][1] = Q[1][3];
            Q[2][3] += m.y()*m.z()-p.y()*p.z(); Q[3][2] = Q[2][3];
        }
       
        //int numRot = 0;   // number of rotations required
        Jacobi.compute(Q, ev, R, 4);
        
        // Each eigenvalue corresponds to the deviation of one structure
        // from the other, after the rotation. The smallest eigenvalue,
        // therefore, must be the best alignment.
       
        
      
        int j = 0;    // generic iterator for eigenvectors/values
        for (int i = 0; i < 4; i++)
            if (ev[i] < ev[j])
                j = i;    // i is a smaller eigenvalue

        // Step 3: Build the rotation matrix from the eigenvector
      
        return buildFromQuaternion(new Tern(R[1][j],R[2][j],R[3][j]),R[0][j]);
    }

    /**
     Concatenate this Transform with another.
     @return <code>this</code> X t2
     */
    public Transform concatenate(Transform t2)
    {
        Transform rt = new Transform(); // holds original values
        
        // Compute 3x3 rotation part
        rt.r00 = r00*t2.r00 + r01*t2.r10 + r02*t2.r20;
        rt.r01 = r00*t2.r01 + r01*t2.r11 + r02*t2.r21;
        rt.r02 = r00*t2.r02 + r01*t2.r12 + r02*t2.r22;
        rt.r10 = r10*t2.r00 + r11*t2.r10 + r12*t2.r20;
        rt.r11 = r10*t2.r01 + r11*t2.r11 + r12*t2.r21;
        rt.r12 = r10*t2.r02 + r11*t2.r12 + r12*t2.r22;
        rt.r20 = r20*t2.r00 + r21*t2.r10 + r22*t2.r20;
        rt.r21 = r20*t2.r01 + r21*t2.r11 + r22*t2.r21;
        rt.r22 = r20*t2.r02 + r21*t2.r12 + r22*t2.r22;
        
        // Compute 1x3 translation part
        rt.tx = tx*t2.r00 + ty*t2.r10 + tz*t2.r20 + t2.tx;
        rt.ty = tx*t2.r01 + ty*t2.r11 + tz*t2.r21 + t2.ty;
        rt.tz = tx*t2.r02 + ty*t2.r12 + tz*t2.r22 + t2.tz;
        return rt;
    }
    
    public void print()
    {
        System.out.printf("%8.5f %8.5f %8.5f\n%8.5f %8.5f %8.5f\n%8.5f %8.5f %8.5f\n%8.5f %8.5f %8.5f\n",
                r00, r01, r02, r10, r11, r12, r20, r21, r22, tx, ty, tz);
    }

}