package model;

import java.io.IOException;

public class CoordSystem {

	/*
	 * x, y, z coordinate system
	 */
	private Tern origin;
	private double[][] orthBase = new double[3][3]; 
	private double[][] transMatrix;
	
	public CoordSystem(Tern origin, Tern baseX, Tern baseY, Tern baseZ)
	{
		this.origin = origin;
		this.orthBase[0][0] = baseX.x();
		this.orthBase[1][0] = baseX.y();
		this.orthBase[2][0] = baseX.z();
		
		this.orthBase[0][1] = baseY.x();
		this.orthBase[1][1] = baseY.y();
		this.orthBase[2][1] = baseY.z();
		
		this.orthBase[0][2] = baseZ.x();
		this.orthBase[1][2] = baseZ.y();
		this.orthBase[2][2] = baseZ.z();
		
		this.transMatrix = MtxInv(orthBase, 3);
	}
	
	public CoordSystem(Residue res)
	{
		Atom n = res.getAtom("N");
		Atom c = res.getAtom("C");
		Atom ca = res.getAtom("CA");
		if(n==null || c == null || ca == null) 
			throw new RuntimeException("atom information insufficient");
		Tern tn = n.getCoordinate();
		Tern tc = c.getCoordinate();
		Tern tca = ca.getCoordinate();
		
		Tern can = tn.subtract(tca);
		can = can.multiply(1.0f/can.length());
		Tern cac = tc.subtract(tca);
		cac = cac.multiply(1.0f/cac.length());
		
		Tern z = can.cross(cac);
		z = z.multiply(1.0f/z.length());
		
		Tern x = can.add(cac);
		x = x.multiply(1.0f/x.length());
		
		Tern y = z.cross(x);
		y = y.multiply(1.0f/y.length());
		
		this.origin = tca;
		this.orthBase[0][0] = x.x();
		this.orthBase[1][0] = x.y();
		this.orthBase[2][0] = x.z();
		
		this.orthBase[0][1] = y.x();
		this.orthBase[1][1] = y.y();
		this.orthBase[2][1] = y.z();
		
		this.orthBase[0][2] = z.x();
		this.orthBase[1][2] = z.y();
		this.orthBase[2][2] = z.z();
		
		this.transMatrix = MtxInv(orthBase, 3);
		
		
	}
	
	public Tern coordTransformation(Tern standardCoord)
	{
		Tern changeOrigin = standardCoord.subtract(this.origin);
		double x = changeOrigin.x()*transMatrix[0][0] + changeOrigin.y()*transMatrix[0][1] + changeOrigin.z()*transMatrix[0][2];
		double y = changeOrigin.x()*transMatrix[1][0] + changeOrigin.y()*transMatrix[1][1] + changeOrigin.z()*transMatrix[1][2];
		double z = changeOrigin.x()*transMatrix[2][0] + changeOrigin.y()*transMatrix[2][1] + changeOrigin.z()*transMatrix[2][2];
		
		return new Tern(x,y,z);
	}
	
	
	
	public static double[][] mtxMultiplication(double[][] a, double[][] b, int n)
	{
		double[][] c = new double[n][n];
		for(int i=0;i<n;i++)
		{
			for(int j=0;j<n;j++)
			{
				for(int k=0;k<n;k++)
				{
					c[i][j] += a[i][k]*b[k][j];
				}
			}
		}
		return c;
	}
	
	public static double[][] MtxInv(double[][] x, int n) {
		  	int i, j, row, col, k;
	        double max, temp;
	        int[] p = new int[n];
	        
	        double[][] a = new double[n][n];
	        double[][] b = new double[n][n];
	        for (i = 0; i < n; i++) {
	            p[i] = i;
	            b[i][i] = 1;
	        }
	        for(i=0;i<n;i++){
	        	for(j=0;j<n;j++)
	        	{
	        		a[i][j] = x[i][j];
	        	}
	        }

	        for (k = 0; k < n; k++) {
	            // find main row
	            max = 0;
	            row = col = i;
	            for (i = k; i < n; i++)
	                for (j = k; j < n; j++) {
	                    temp = Math.abs(b[i][j]);
	                    if (max < temp) {
	                        max = temp;
	                        row = i;
	                        col = j;
	                    }
	                }
	            // row swap, put main row to k row
	            if (row != k) {
	                for (j = 0; j < n; j++) {
	                    temp = a[row][j];
	                    a[row][j] = a[k][j];
	                    a[k][j] = temp;
	                    temp = b[row][j];
	                    b[row][j] = b[k][j];
	                    b[k][j] = temp;
	                }
	                i = p[row];
	                p[row] = p[k];
	                p[k] = i;
	            }
	            if (col != k) {
	                for (i = 0; i < n; i++) {
	                    temp = a[i][col];
	                    a[i][col] = a[i][k];
	                    a[i][k] = temp;
	                }
	            }
	            
	            for (j = k + 1; j < n; j++)
	                a[k][j] /= a[k][k];
	            for (j = 0; j < n; j++)
	                b[k][j] /= a[k][k];
	            a[k][k] = 1;

	            for (j = k + 1; j < n; j++) {
	                for (i = 0; i < k; i++)
	                    a[i][j] -= a[i][k] * a[k][j];
	                for (i = k + 1; i < n; i++)
	                    a[i][j] -= a[i][k] * a[k][j];
	            }
	            for (j = 0; j < n; j++) {
	                for (i = 0; i < k; i++)
	                    b[i][j] -= a[i][k] * b[k][j];
	                for (i = k + 1; i < n; i++)
	                    b[i][j] -= a[i][k] * b[k][j];
	            }
	            for (i = 0; i < k; i++)
	                a[i][k] = 0;
	            a[k][k] = 1;
	        }
	        
	        return b;
	    }
	
	public static void main(String[] args) throws IOException
	{
		PDB pdb = new PDB("F:\\1UBQ.pdb");
		Residue res = pdb.getFirstChain().getResidue(8);
		CoordSystem cs = new CoordSystem(res);
		for(String s : res.getAtoms().keySet())
		{
			Atom a = res.getAtom(s);
			Tern ta = a.getCoordinate();
			Tern localTa = cs.coordTransformation(ta);
			System.out.printf("%-3s %6.2f %6.2f %6.2f\n",s,localTa.x(),localTa.y(),localTa.z());
		//	System.out.println(localTa.toString());
		}
		
	}
}
