package model;

import java.io.IOException;
import java.io.Serializable;
import java.util.*;

import similarSearch.PairPara;

public class InterfacePair implements Serializable
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 8234683490586157884L;
	public final InterfaceResidue irA;
	public final InterfaceResidue irB;
	public final ArrayList<Tern> eightAtoms;
	
	public final boolean sameChain;
	
	public InterfacePair(InterfaceResidue irA, InterfaceResidue irB)
	{
		this.irA = irA;
		this.irB = irB;
		Residue resA = irA.getResidue();
		Residue resB = irB.getResidue();
		
        if(!resA.infoSufficient() || !resB.infoSufficient())
        {
                throw new RuntimeException("residue info not sufficient");
        }
        ArrayList<Tern> resAFourAtoms = resA.getBackboneFourAtomTerns();
        ArrayList<Tern> resBFourAtoms = resB.getBackboneFourAtomTerns();
        this.eightAtoms = new ArrayList<Tern>();
        for(int i=0;i<4;i++)
        {
                eightAtoms.add(resAFourAtoms.get(i));
        }
        for(int i=0;i<4;i++)
        {
                eightAtoms.add(resBFourAtoms.get(i));
        }
        if(resA.getChainID() != resB.getChainID())
        	sameChain = false;
        else 
        	sameChain = true;
	}

	public InterfacePair(String line)
	{
		if(line.length()!= 303)
			throw new RuntimeException("line length error");
		this.irA = new InterfaceResidue(line.substring(0, 55));
		this.irB = new InterfaceResidue(line.substring(56,111));
		Scanner sc = new Scanner(line.substring(112,line.length()));
		this.eightAtoms = new ArrayList<Tern>();
		
		for(int i=0;i<8;i++)
		{
			float x = sc.nextFloat();
			float y = sc.nextFloat();
			float z = sc.nextFloat();
			this.eightAtoms.add(new Tern(x,y,z));
		}
		sc.close();
		if(irA.getChainID() != irB.getChainID())
			this.sameChain = false;
		else 
			this.sameChain = true;
		
	}
	
	public boolean similarTo(InterfacePair other, PairPara ppr)
	{
		if(this.sameChain && !other.sameChain || !this.sameChain && other.sameChain) return false;
		if(this.sameChain && other.sameChain)
		{
			if(this.seqDistance() != other.seqDistance()) return false;
		}
		if(!this.irA.similarTo(other.irA, ppr.resParaA())) return false;
		if(!this.irB.similarTo(other.irB, ppr.resParaA())) return false;
		
		double rmsd = math.RMSD.rmsd(this.eightAtoms, other.eightAtoms);
		if(rmsd > ppr.cutoff()) return false;
		
		return true;
	}
	
	public int seqDistance()
	{
		int thisSeqDist = Math.abs(this.irA.getResidue().getID() - this.irB.getResidue().getID());
		if(thisSeqDist > 5) thisSeqDist = 5;
		return  thisSeqDist;
	}
	
	public InterfacePair mirror()
	{
		return new InterfacePair(this.irB, this.irA);
	}
	
	public String toString()
	{
		String line = this.irA.toString() + " " + this.irB.toString();
        for(int i=0;i<8;i++)
        {
                line=line + String.format("%8.2f", this.eightAtoms.get(i).x());
                line=line + String.format("%8.2f", this.eightAtoms.get(i).y());
                line=line + String.format("%8.2f", this.eightAtoms.get(i).z());
        }
        return line;
	}
	
	public static void main(String[] args) throws IOException
	{
		String s = "1yjg B  116 VAL1000 T  -74.40  137.52 0.972 0.255 4.678 1yjg D  199 ASN3200 H  -71.81  -20.25 0.891 0.469 4.739  29.52   3.72   8.70  29.17   2.99   7.36  28.88   3.07   9.85  29.14   5.19   8.59  24.77  -1.76   9.05  24.13  -0.36   9.16  25.11  -2.03   7.66  23.82  -2.79   9.67";
		InterfacePair ip = new InterfacePair(s);
		String ss2 = ip.toString();
		if(s.equals(ss2))
			System.out.println("ok");
		else
			System.out.println("err");
	}
}
