package model;

import java.util.*;

public class ProteinChain 
{
	private String pdbId;
	private char chainId;
	private ArrayList<Residue> residues=new ArrayList<Residue>();
	private ArrayList<SideChain> sideChains = new ArrayList<SideChain>();
	private HashMap<Integer,Double> omegas = new HashMap<Integer,Double>();
	
	public ProteinChain()
	{	
	}
	
	public ProteinChain(String pdbId,char chainId)
	{
		this.pdbId=pdbId;
		this.chainId=chainId;
		if(this.chainId==' ')
			this.chainId='-';
	}
	
	public ProteinChain(String pdbNAME)
	{
		this.pdbId=pdbNAME;
		this.chainId='-';
	}

	public String getPDBId()
	{
		return this.pdbId;
	}
	
	public char getChainId()
	{
		return this.chainId;
	}
	
	public int getChainLength()
	{
		return this.residues.size();
	}
	
	public ArrayList<Residue> getResidue()
	{
		return this.residues;
	}
	
	public Residue getResidue(int resId)
	{
		for(Residue res : residues)
		{
			if(res.getID() == resId)
			{
				return res;
			}
		}
		return new Residue();
	}
	
	
	
	public void addResidue(Residue res)
	{
		int residueNumber = this.getResidue().size();
		res.setResIDSD(residueNumber);
		residues.add(res);
	}
	
	public void computeSideChain()
	{
		this.sideChains = new ArrayList<SideChain>();
		for(Residue res : this.residues)
		{
			if(!ResName.isResName(res.getName()))
				this.sideChains.add(null);
			else
				this.sideChains.add(new SideChain(res));
		}
	}
	
	public ArrayList<ResPair> findContactPairs(double cutoff)
	{
		ArrayList<ResPair> pairs=new ArrayList<ResPair>();
		ArrayList<Residue> residues=this.getResidue();
		for(int i=0;i<residues.size();i++)
		{
			for(int j=i+1;j<residues.size();j++)
			{
				Residue resA=residues.get(i);
				Residue resB=residues.get(j);
				if(resA.infoSufficient()&&resB.infoSufficient())
				{
					Tern eCogA=resA.getExpectedCOG();
					Tern eCogB=resB.getExpectedCOG();
					double expCogDistance=eCogA.distance(eCogB);
					if(expCogDistance<cutoff)
					{
						pairs.add(new ResPair(resA,resB,this.pdbId,this.chainId));
					}
				}
			}
		}
		return pairs;
	}
	
	public ArrayList<RotPair> findContactRotPairs(double cutoff)
	{
		ArrayList<RotPair> pairs=new ArrayList<RotPair>();
		ArrayList<Residue> residues=this.getResidue();
		for(int i=0;i<residues.size();i++)
		{
			for(int j=i+1;j<residues.size();j++)
			{
				Residue resA=residues.get(i);
				Residue resB=residues.get(j);
				if(resA.infoSufficient()&&resB.infoSufficient())
				{
					Tern eCogA=resA.getExpectedCOG();
					Tern eCogB=resB.getExpectedCOG();
					double expCogDistance=eCogA.distance(eCogB);
					if(expCogDistance<cutoff)
					{
						pairs.add(new RotPair(resA,resB,this.pdbId,this.chainId));
					}
				}
			}
		}
		return pairs;
	}
	
	public boolean checkBackbone()
	{
		int resNum = this.getChainLength();
		if(resNum < 2)
			return true;
		for(Residue res : this.residues)
		{
			if(!res.infoSufficient())
				return false;
		}
		
		Atom N = this.residues.get(0).getAtom("N");
		Atom CA = this.residues.get(0).getAtom("CA");
		double lenA = N.distance(CA);
		if(lenA > 2.0)
		{
//			System.out.println("resID : 0");
			return false;
		}
		Atom C = this.residues.get(0).getAtom("C");
		double lenB = CA.distance(C);
		if(lenB > 2.0)
		{	
//			System.out.println("resID : 0");
			return false;
		}
		for(int i=1;i<this.getChainLength();i++)
		{
			N = this.residues.get(i).getAtom("N");
			double lenC = C.distance(N);
			CA = this.residues.get(i).getAtom("CA");
			lenA = N.distance(CA);
			C = this.residues.get(i).getAtom("C");
			lenB = CA.distance(C);
			if(lenC > 2.0 || lenA > 2.0 || lenB > 2.0)
			{
//				System.out.println("resID : " + i);
				return false;
			}
		}
		return true;
	}
	
	public String getSequence()
	{
		String seq = "";
		for(Residue r : this.residues)
		{
			seq = seq + ResName.triToSin(r.getName());
		}
		return seq;
	}
	
	public String getRotamerSeq()
	{
		String seq = "";
		for(SideChain sdc : this.sideChains)
		{
			seq = seq + sdc.getDunbrackCode();
		}
		return seq;
	}
	
	public ArrayList<Tern> getMainchainAtoms()
	{
		//CA, N, C, O
		ArrayList<Tern> thisMainchainAtoms = new ArrayList<Tern>();
		for(Residue res : this.residues)
		{
			ArrayList<Tern> fourHeavyAtoms = res.getMainchainAtomTerns();
			if(fourHeavyAtoms == null)
				throw new RuntimeException("pdb info insufficient: " + this.pdbId + " " + this.chainId + " " + res.getID());
			for(int i=0;i<fourHeavyAtoms.size();i++)
			{
				thisMainchainAtoms.add(fourHeavyAtoms.get(i));
			}
		}
		return thisMainchainAtoms;
	}
	
	public double mainchainRmsd(ProteinChain other)
	{
		//rmsd of CA, N, C, O
		ArrayList<Tern> thisMainchainAtoms = this.getMainchainAtoms();
		ArrayList<Tern> otherMainchainAtoms = other.getMainchainAtoms();
		if(this.getChainLength() != other.getChainLength())
			throw new RuntimeException("chain length not equal");
		return math.RMSD.rmsd(thisMainchainAtoms, otherMainchainAtoms); 
	}
	
	
	public double selectResiduesMainchainRmsd(ProteinChain other, ArrayList<Integer> selectResidueIDSD)
	{
		ArrayList<Tern> thisMainchainAtoms = new ArrayList<Tern>();
		ArrayList<Tern> otherMainchainAtoms = new ArrayList<Tern>();
		if(this.getChainLength() != other.getChainLength())
			throw new RuntimeException("chain length not equal");
		for(int k=0;k<selectResidueIDSD.size();k++)
		{
			Residue res = this.getResidue().get(k);
			ArrayList<Tern> fourHeavyAtoms = res.getMainchainAtomTerns();
			if(fourHeavyAtoms == null)
				throw new RuntimeException("pdb info insufficient: " + this.pdbId + " " + this.chainId + " " + res.getID());
			for(int i=0;i<fourHeavyAtoms.size();i++)
			{
				thisMainchainAtoms.add(fourHeavyAtoms.get(i));
			}
		}
		for(int k=0;k<selectResidueIDSD.size();k++)
		{
			Residue res = other.getResidue().get(k);
			ArrayList<Tern> fourHeavyAtoms = res.getMainchainAtomTerns();
			if(fourHeavyAtoms == null)
				throw new RuntimeException("pdb info insufficient: " + this.pdbId + " " + this.chainId + " " + res.getID());
			for(int i=0;i<fourHeavyAtoms.size();i++)
			{
				otherMainchainAtoms.add(fourHeavyAtoms.get(i));
			}
		}
		return math.RMSD.rmsd(thisMainchainAtoms, otherMainchainAtoms); 
	}
	
	public void computeOmega()
	{
		this.omegas.put(this.residues.get(0).getID(), 180.0);
		for(int i=1;i<this.residues.size();i++)
		{
			int resID = this.residues.get(i).getID();
			double omg = omega(this.residues.get(i-1),this.residues.get(i));
			this.omegas.put(resID, omg);
		}
	}
	
	public double getOMG(int resID)
	{
		return this.omegas.get(resID);
	}
	
	public static double omega(Residue resFormer, Residue resCurrent)
	{
		if(!resFormer.getAtoms().containsKey("CA") || !resFormer.getAtoms().containsKey("C") || !resCurrent.getAtoms().containsKey("CA") || !resCurrent.getAtoms().containsKey("N"))
		{
			System.err.println("Warning: lack atom to compute omega, " + resCurrent.getID());
			return 180.0;
		}
		Atom a = resFormer.getAtom("CA");
		Atom b = resFormer.getAtom("C");
		Atom c = resCurrent.getAtom("N");
		Atom d = resCurrent.getAtom("CA");
		if(b.distance(c) > 2.0)
		{
			System.err.println("Warning: snap peptide " + resCurrent.getID());
			return 180.0;
		}
		double omega = math.Dihedral.dihedral(a.getCoordinate(), b.getCoordinate(), c.getCoordinate(), d.getCoordinate());
		if(omega < -90.0)
			omega = omega+360.0;
		return omega;
	}
}
