package model;
import java.io.*;

import similarSearch.*;

public class ResPair 
{
	private Residue resA;
	private Residue resB;
	
	private String pdbID="xxxx";
	private char chainID='-';
	
	private double CaCaDistance;
	private double angleA;
	private double angleB;
	private double dihedral;
	
	private double expCogDistance;
	private int seqDistance;
	
	public StrideRes strResA;
	public StrideRes strResB;
	public boolean hasStrInfo=false;
	
	public ResPair() {} 
	
	public ResPair(Residue resA,Residue resB)
	{
		this.resA=resA;
		this.resB=resB;
		int resAid=this.resA.getID();
		int resBid=this.resB.getID();
		this.seqDistance=Math.abs(resAid-resBid);
		
		Tern resACa=this.resA.getAtom("CA").getCoordinate();
		if(this.resB.getAtom("CA")==null)
		{
			throw new RuntimeException("can't find atom "+this.resB.getIDSD()+" CA");
		}
		
		Tern resBCa=this.resB.getAtom("CA").getCoordinate();
		Tern CaCaAB=resBCa.subtract(resACa);
		Tern CaCaBA=resACa.subtract(resBCa);
		this.CaCaDistance=resACa.distance(resBCa);
		this.angleA=CaCaAB.angle(this.resA.getCaCbVector());
		this.angleB=CaCaBA.angle(this.resB.getCaCbVector());
		Tern n1=CaCaAB.cross(this.resA.getCaCbVector()).cross(CaCaAB);
		Tern n2=CaCaBA.cross(this.resB.getCaCbVector()).cross(CaCaBA);
		this.dihedral=n1.angle(n2);
		if(this.dihedral!=0)
		{
			double symbol=n1.cross(n2).dotProduct(CaCaBA);
			symbol=symbol/Math.abs(symbol);
			this.dihedral=this.dihedral*symbol;
		}
		
		this.expCogDistance=resA.getExpectedCOG().distance(resB.getExpectedCOG());
		this.seqDistance=resB.getID()-resA.getID();
	}
	
	public ResPair(Residue resA,Residue resB,String pdbID,char chainID)
	{
		this(resA,resB);
		this.pdbID=pdbID;
		this.chainID=chainID;
	}
	
	public ResPair(Residue resA, Residue resB, StrideRes strA, StrideRes strB, int seqDist)
	{
		this.resA = resA;
		this.resB = resB;
		this.strResA = strA;
		this.strResB = strB;
		this.seqDistance = seqDist;
		this.hasStrInfo = true;
	}
	
	
	public ResPair(String line)
	{
		this.pdbID=line.substring(0,4);
		this.chainID=line.charAt(5);
		this.resA=new Residue(line.substring(13,16),Integer.parseInt(line.substring(8,12).trim()));
		this.resB=new Residue(line.substring(22,25),Integer.parseInt(line.substring(17,21).trim()));
		
		this.CaCaDistance=Double.parseDouble(line.substring(28,34).trim());
		this.angleA=Double.parseDouble(line.substring(36,43).trim());
		this.angleB=Double.parseDouble(line.substring(45,52).trim());
		this.dihedral=Double.parseDouble(line.substring(53,61).trim());
		this.expCogDistance=Double.parseDouble(line.substring(63,70).trim());
		
		char ssTypeA=line.charAt(71);
		float phiA=Float.parseFloat(line.substring(73,81).trim());
		float psiA=Float.parseFloat(line.substring(82,90).trim());
		float areaA=Float.parseFloat(line.substring(92,97).trim());
		char ssTypeB=line.charAt(98);
		float phiB=Float.parseFloat(line.substring(100,108).trim());
		float psiB=Float.parseFloat(line.substring(109,117).trim());
		float areaB=Float.parseFloat(line.substring(119,124).trim());
		this.strResA=new StrideRes(this.pdbID,this.chainID,this.resA.getID(),ssTypeA,phiA,psiA,areaA);
		this.strResB=new StrideRes(this.pdbID,this.chainID,this.resB.getID(),ssTypeB,phiB,psiB,areaB);
		
		this.seqDistance=this.resB.getID()-this.resA.getID();
	}
	
	public PairRelativePosition relativePosition()
	{
		return new PairRelativePosition(this.CaCaDistance,this.angleA,this.angleB,this.dihedral);
	}
	
	public Residue getResA()
	{
		return this.resA;
	}
	
	public void setChainID(char c)
	{
		this.chainID = c;
	}
	
	public Residue getResB()
	{
		return this.resB;
	}
	
	public void addStrInfo(StrideProtein sp)
	{
		StrideRes sResA=sp.getStrideResidue(""+chainID+resA.getID());
		StrideRes sResB=sp.getStrideResidue(""+chainID+resB.getID());
		this.strResA=sResA;
		this.strResB=sResB;
		if(sResA!=null && sResB!=null)
		{
			this.hasStrInfo=true;
		}
		return;

	}
	
	public ResPair mirror()
	{
		ResPair mirrorPair=new ResPair();
		mirrorPair.resA=this.resB;
		mirrorPair.resB=this.resA;
		mirrorPair.angleA=this.angleB;
		mirrorPair.angleB=this.angleA;
		mirrorPair.CaCaDistance=this.CaCaDistance;
		mirrorPair.dihedral=this.dihedral;
		mirrorPair.chainID=this.chainID;
		mirrorPair.expCogDistance=this.CaCaDistance;
		mirrorPair.hasStrInfo=this.hasStrInfo;
		mirrorPair.pdbID=this.pdbID;
		mirrorPair.seqDistance=this.seqDistance;
		mirrorPair.strResA=this.strResB;
		mirrorPair.strResB=this.strResA;
		return mirrorPair;
	}
	
	public boolean hasStrideInfo()
	{
		return this.hasStrInfo;
	}
	
	public int seqDistance()
	{
		return this.seqDistance;
	}
	
	public double expCogDistance()
	{
		return this.expCogDistance;
	}
	
	public String toString()
	{

		String pdbIDFormat=String.format("%4s", this.pdbID);
		String line=pdbIDFormat+" "+this.chainID+" ";
		line=line+String.format("%5d", this.resA.getIDSD());
		line=line+" "+resA.getName();
		line=line+String.format("%5d", this.resB.getIDSD());
		line=line+" "+resB.getName();
		line=line+String.format("%9.3f", this.CaCaDistance);
		line=line+String.format("%9.3f", this.angleA);
		line=line+String.format("%9.3f", this.angleB);
		line=line+String.format("%9.3f", this.dihedral);
		line=line+String.format("%9.3f", this.expCogDistance);
		if(this.hasStrInfo)
		{
			line=line+this.strResA.toString()+this.strResB.toString();
		}
		return line;
	}

	public static void main(String[] args) throws IOException
	{
		PDB pdb=new PDB("E:\\pairwiseDesign\\pdb\\0\\pdb1b0n.ent","1b0n");
		StrideProtein sp=new StrideProtein("E:\\pairwiseDesign\\meaPDBStride\\0\\1b0n.str");
		Residue resA=pdb.getChain('A').getResidue(6);
		Residue resB=pdb.getChain('A').getResidue(61);
		ResPair rp=new ResPair(resA,resB,"1b0n",'A');
		rp.addStrInfo(sp);
		String s=rp.toString();
		System.out.println(s);
		double eD=rp.expCogDistance();
		System.out.println(eD);
		
		ResPair rp2=new ResPair(s);
		String s2=rp2.toString();
		System.out.println(s2);
	}
}
