package model;

import java.io.IOException;
import java.io.Serializable;
import java.util.*;

import burial.Rotation;
import math.Matrix;
import math.Transform;

public class Residue implements Serializable
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 3140710021878561200L;
	private HashMap<String,Atom> atoms=new HashMap<String,Atom>();
	private int resID;
	private int resIDSD=-1;
	private String triName;
	private char chainID = '-';
	
	public Residue()
	{
		this.resID=0;
		this.triName="";
	}
	
	public Residue(String s,int i)
	{
		this.resID=i;
		this.triName=s;
	}
	
	public String getName()
	{
		return this.triName;
	}

	public int getID()
	{
		return this.resID;
	}
	
	public int getIDSD()
	{
		return this.resIDSD;
	}
	
	public void setName(String s)
	{
		this.triName=s;
	}
	
	public void setResID(int i)
	{
		this.resID=i;
	}
	
	public void setChainID(char c)
	{
		this.chainID = c;
	}
	
	public char getChainID()
	{
		return this.chainID;
	}
	
	public void addAtom(Atom a)
	{
		String atomName=a.getName();
		this.atoms.put(atomName, a);
	}
	
	public void setResIDSD(int resIDStandardized)
	{
		this.resIDSD=resIDStandardized;
	}
	
	public static boolean isBackbone(Atom a)
	{
		String atomName=a.getName();
		HashSet<String> backBoneAtomName=new HashSet<String>();
		backBoneAtomName.add("C");
		backBoneAtomName.add("CA");
		backBoneAtomName.add("N");
		backBoneAtomName.add("O");
		backBoneAtomName.add("H");
		backBoneAtomName.add("H2");
		backBoneAtomName.add("H3");
		backBoneAtomName.add("HA");
		backBoneAtomName.add("1HA");
		backBoneAtomName.add("2HA");
		backBoneAtomName.add("OXT");
		backBoneAtomName.add("HXT");
		
		return backBoneAtomName.contains(atomName);		
	}
	
	
	public HashMap<String,Atom> getAtoms()
	{
		return this.atoms;
	}
	
	public Atom getAtom(String name)
	{
		if(this.atoms.containsKey(name))
		{
			return this.atoms.get(name);
		}
		return null;	
	}
	
	public ArrayList<Atom> getBackbone()
	{
		
		ArrayList<Atom> backBone=new ArrayList<Atom>();
		for(Atom a : this.atoms.values())
		{
			if(isBackbone(a))
			{
				backBone.add(a);
			}
		}
		return backBone;
	}
	
	public ArrayList<Atom> getSidechain()
	{
		ArrayList<Atom> sideChain=new ArrayList<Atom>();
		for(Atom a : this.atoms.values())
		{
			if(!isBackbone(a))
			{
				sideChain.add(a);
			}
		}
		return sideChain;
	}
	
	public Tern getSidechainCOG()
	{
		ArrayList<Atom> sideChain=new ArrayList<Atom>();
		sideChain=this.getSidechain();
		
		Tern COG=new Tern(0,0,0);
		float weight=0;
		for(Atom a : sideChain)
		{
			weight=weight+a.atomWeight();
			COG=COG.add(a.getCoordinate().multiply(a.atomWeight())); 
		}
		
		if(weight == 0)
		{
			return null;
		}
		COG=COG.multiply(1/weight);
		return COG;
	}


	public Tern getCaCbVector()
	{
		if(!atoms.containsKey("CB"))
		{
			return this.getCaCbVectorOfGly();
		}
		Atom ca = this.getAtom("CA");
		Atom cb = this.getAtom("CB");
		
		Tern cacoo=ca.getCoordinate();
		Tern cbcoo=cb.getCoordinate();
		
		return cbcoo.subtract(cacoo);
		
	}

	public Tern getCaCbVectorOfGly()
	{
		if(!this.cbInfoSufficient())
		{
			System.err.println("Warning: residue information insufficient, residue ID " + this.resID);
			return null;
		}
		Atom ca=this.getAtom("CA");
		Atom c=this.getAtom("C");
		Atom n=this.getAtom("N");
		Tern cac=c.getCoordinate().subtract(ca.getCoordinate());
		Tern can=n.getCoordinate().subtract(ca.getCoordinate());
		cac=cac.multiply(1/cac.length());
		can=can.multiply(1/can.length());
		Tern lawOfCaCN=can.cross(cac);
		Tern CaCb=cac.add(can).multiply(-0.5f).add(lawOfCaCN.multiply(0.816497f/lawOfCaCN.length()));
		CaCb=CaCb.multiply(1.54f/CaCb.length());
		return CaCb;                                                                                                                                                                                                                                                                                                                                                                                                                                                         
	}
	
	public ArrayList<Tern> getBackboneFourAtomTerns()
	{
		//order: Ca, Cb, N, C
		if(!this.infoSufficient())
			return null;
		ArrayList<Tern> fourTerns = new ArrayList<Tern>();
		Atom ca=this.getAtom("CA");
		
		Atom n=this.getAtom("N");
		Atom c=this.getAtom("C");
		fourTerns.add(ca.getCoordinate());
		fourTerns.add(ca.getCoordinate().add(getCaCbVector()));
		fourTerns.add(n.getCoordinate());
		fourTerns.add(c.getCoordinate());
		return fourTerns;
	}
	
	public ArrayList<Tern> getMainchainAtomTerns()
	{
		//order: Ca, N, C, O
		if(!this.infoSufficient())
			return null;
		ArrayList<Tern> terns = new ArrayList<Tern>();
		Atom ca=this.getAtom("CA");
		Atom n=this.getAtom("N");
		Atom c=this.getAtom("C");
		terns.add(ca.getCoordinate());	
		terns.add(n.getCoordinate());
		terns.add(c.getCoordinate());
		if(this.atoms.containsKey("O"))
			terns.add(this.getAtom("O").getCoordinate());
		else if(this.atoms.containsKey("OT1")) terns.add(this.getAtom("OT1").getCoordinate());
		else if(this.atoms.containsKey("OT2")) terns.add(this.getAtom("OT2").getCoordinate());
		return terns;
	}
	
	public boolean infoSufficient()
	{
		boolean suf1=atoms.containsKey("CA")&&atoms.containsKey("N")&&atoms.containsKey("C")&&atoms.containsKey("O");
		boolean suf2=atoms.containsKey("CA")&&atoms.containsKey("N")&&atoms.containsKey("C")&&atoms.containsKey("OT1")&&atoms.containsKey("OT2");
		
		return suf1 || suf2;
	}
	
	public boolean cbInfoSufficient()
	{
		return atoms.containsKey("CA")&&atoms.containsKey("N")&&atoms.containsKey("C");
	}
	
	public Tern getExpectedCOG()
	{
		Tern CaCbVector;
		CaCbVector=this.getCaCbVector();
		Tern eCOG=this.getAtom("CA").getCoordinate().add(CaCbVector.multiply(2.549f/CaCbVector.length()));
		return eCOG;		
	}
	
	public Tern getExpectedCOG2()
	{
		Tern[] ternsA =  new Tern[3];
		ternsA[0] = atoms.get("CA").getCoordinate();
		ternsA[1] = atoms.get("N").getCoordinate();
		ternsA[2] = atoms.get("C").getCoordinate();
		
		Tern cogA = ternsA[0].add(ternsA[1]).add(ternsA[2]).divide(3.0f);
		
		Tern[] ternsA2 = new Tern[3];
		ternsA2[0] = ternsA[0].subtract(cogA);
		ternsA2[1] = ternsA[1].subtract(cogA);
		ternsA2[2] = ternsA[2].subtract(cogA);
		

		Tern[] ternsB = new Tern[3];
		ternsB[0] = new Tern(0.45367f,-0.27067f,0.19267f);
		ternsB[1] = new Tern(-0.84133f,-0.28867f,0.88367f);
		ternsB[2] = new Tern(0.38767f,0.55933f,-1.07633f);
		
		Tern sidechainEcogB = new Tern(1.48345f,-2.42159f,0.14673f);
		Transform tf =  Transform.buildRotation(ternsA2,ternsB);
		Tern sidechainEcogA = tf.transform(sidechainEcogB);
		
		sidechainEcogA = sidechainEcogA.add(cogA);

		return sidechainEcogA;
	}
	
	public ArrayList<Atom> getMEAAtoms()
	{
		if(!this.cbInfoSufficient())
		{
			System.err.println("Warning: residue information insufficient, residue ID " + this.resID);
			return null;
		}
		Tern caCbVeTern = this.getCaCbVectorOfGly();
		ArrayList<Atom> atoms = new ArrayList<Atom>();
		Atom ca=this.getAtom("CA");
		Atom cb = new Atom("CB",ca.getCoordinate().add(caCbVeTern));
		Atom cg = new Atom("CG",cb.getCoordinate().add(caCbVeTern));
		Atom cd = new Atom("CD",cg.getCoordinate().add(caCbVeTern));
		
		atoms.add(cb);
		atoms.add(cg);
		atoms.add(cd);
		return atoms;
	}
	
	public ArrayList<Atom> getMEDAtoms()
	{
		if(!this.cbInfoSufficient())
		{
			System.err.println("Warning: residue information insufficient, residue ID " + this.resID);
			return null;
		}
		Atom ca=this.getAtom("CA");
		Atom c=this.getAtom("C");
		Atom n=this.getAtom("N");
		Tern cac=c.getCoordinate().subtract(ca.getCoordinate());
		Tern can=n.getCoordinate().subtract(ca.getCoordinate());
		cac=cac.multiply(1/cac.length());
		can=can.multiply(1/can.length());
		Tern lawOfCaCN=can.cross(cac);
		Tern CaCb=cac.add(can).multiply(-0.5f).add(lawOfCaCN.multiply(0.816497f/lawOfCaCN.length()));
		
		float ccLen = 2.0f;
		Tern CaCp = CaCb.multiply(ccLen);
		Tern C2Cq = CaCb.multiply(ccLen);

		
		Tern CaCN = cac.add(can);
		Tern OC2 = CaCN.subtract(CaCb.multiply(CaCN.dotProduct(CaCb)/CaCb.length()));
		OC2 = OC2.divide(OC2.length());
		Tern CpC2 = CaCb.multiply(ccLen/3.0f).subtract(OC2.multiply(ccLen*0.94281f));
		
		
		ArrayList<Atom> atoms= new ArrayList<Atom>();
		Tern cp = ca.getCoordinate().add(CaCp);
		Tern c2 = cp.add(CpC2);
		Tern cq = c2.add(C2Cq);

		atoms.add(new Atom("CB",cp));
		
		atoms.add(new Atom("CG",c2));
		atoms.add(new Atom("CD",cq));
		
		return atoms;
	}
	
	public static Matrix getCAMatrix(Residue res)
	{
		Tern C = res.getAtom("C").getCoordinate();
		Tern N = res.getAtom("N").getCoordinate();
		Tern CA = res.getAtom("CA").getCoordinate();
		
		Tern NCA = CA.subtract(N);
		Tern x = NCA.divide(NCA.length());
		
	//	System.out.println("NCA " + NCA.toString());
		Tern CAC = C.subtract(CA);
		Tern z = CAC.cross(NCA);
		z = z.divide(z.length());
		
		Tern y = z.cross(x);
		
		double[][] values = {
				{x.x(),y.x(),z.x()},
				{x.y(),y.y(),z.y()},
				{x.z(),y.z(),z.z()}
		};
		Matrix m = new Matrix(values);
		return m;
	}
	
	public ArrayList<Atom> getPSDAtoms()
	{
		float d1 = 1.7f;
		float d2 = 1.5f;
		float chi1 = -90.0f;
		
		if(!this.cbInfoSufficient()) return null;
		double dihed = 122.50;
		double ang = 69.68;
		Matrix mx = Rotation.rotByX(dihed/180.0*Math.PI);
		Matrix mz = Rotation.rotByZ(ang/180.0*Math.PI);
		Matrix m = getCAMatrix(this).multiply(mx).multiply(mz);
		Tern cbInner = new Tern(d1,0,0);
		Tern cb = Tern.transformation(m, cbInner).add(this.getAtom("CA").getCoordinate());
		Atom CB = new Atom("CB", cb);
		
		double ang2 = 70.5;
		Matrix mx2 = Rotation.rotByX(-1.0*chi1/180.0*Math.PI);
		Matrix mz2 = Rotation.rotByZ(ang2/180.0*Math.PI);
		Matrix m2 = m.multiply(mx2).multiply(mz2);
		Tern cgInner = new Tern(d2,0,0);
		Tern cg = Tern.transformation(m2, cgInner).add(cb);
		Atom CG = new Atom("CG",cg);
		
		double chi2 = 180;
		double ang3 = 70.5;
		Matrix mx3 = Rotation.rotByX(-1.0*chi2/180.0*Math.PI);
		Matrix mz3 = Rotation.rotByZ(ang3/180.0*Math.PI);
		Matrix m3 = m2.multiply(mx3).multiply(mz3);
		Tern cdInner = new Tern(d1,0,0);
		Tern cd = Tern.transformation(m3, cdInner).add(cg);
		Atom CD = new Atom("CD",cd);
		
		ArrayList<Atom> list = new ArrayList<Atom>();
		list.add(CB);
		list.add(CG);
		list.add(CD);
		return list;
	}

	/*
	public ArrayList<Atom> getPSD2Atoms()
	{
		double d1 = 2.0;
		double d2 = 2.0;
		double chi1 = -100;
		
		if(!this.cbInfoSufficient()) return null;
		double dihed = 122.50;
		double ang = 69.68;
		Matrix mx = Rotation.rotByX(dihed/180.0*Math.PI);
		Matrix mz = Rotation.rotByZ(ang/180.0*Math.PI);
		Matrix m = getCAMatrix(this).multiply(mx).multiply(mz);
		Tern cbInner = new Tern(d1,0,0);
		Tern cb = Tern.transformation(m, cbInner).add(this.getAtom("CA").getCoordinate());
		Atom CB = new Atom("CB", cb);
		
		double ang2 = 70.5;
		Matrix mx2 = Rotation.rotByX(-1.0*chi1/180.0*Math.PI);
		Matrix mz2 = Rotation.rotByZ(ang2/180.0*Math.PI);
		Matrix m2 = m.multiply(mx2).multiply(mz2);
		Tern cgInner = new Tern(d2,0,0);
		Tern cg = Tern.transformation(m2, cgInner).add(cb);
		Atom CG = new Atom("CG",cg);
		
		double chi2 = 180;
		double ang3 = 70.5;
		Matrix mx3 = Rotation.rotByX(-1.0*chi2/180.0*Math.PI);
		Matrix mz3 = Rotation.rotByZ(ang3/180.0*Math.PI);
		Matrix m3 = m2.multiply(mx3).multiply(mz3);
		Tern cdInner = new Tern(d1,0,0);
		Tern cd = Tern.transformation(m3, cdInner).add(cg);
		Atom CD = new Atom("CD",cd);
		
		ArrayList<Atom> list = new ArrayList<Atom>();
		list.add(CB);
		list.add(CG);
		list.add(CD);
		return list;
	}
	*/
	
	public ArrayList<Atom> getPSEAtoms()
	{
		float d1 = 2.6f;
		float d2 = 1.5f;
		float d3 = 2.2f;
		float chi = 170f;
		
		if(!this.cbInfoSufficient()) return null;
		float dihed = 122.50f;
		float ang = 69.68f;
		Matrix mx = Rotation.rotByX(dihed/180.0*Math.PI);
		Matrix mz = Rotation.rotByZ(ang/180.0*Math.PI);
		Matrix m = getCAMatrix(this).multiply(mx).multiply(mz);
		Tern cbInner = new Tern(d1,0,0);
		Tern cb = Tern.transformation(m, cbInner).add(this.getAtom("CA").getCoordinate());
		Atom CB = new Atom("CB", cb);
		
		double ang2 = 70.5;
		Matrix mx21 = Rotation.rotByX(-1.0*chi/180.0*Math.PI);
		Matrix mz2 = Rotation.rotByZ(ang2/180.0*Math.PI);
		Matrix m21 = m.multiply(mx21).multiply(mz2);
		Tern cgInner1 = new Tern(d2,0,0);
		Tern cg1 = Tern.transformation(m21, cgInner1).add(cb);
		Atom CG1 = new Atom("CG1",cg1);
		
		Matrix mx22 = Rotation.rotByX(-1.0*(chi + 120)/180.0*Math.PI);
		Matrix m22 = m.multiply(mx22).multiply(mz2);
		Tern cgInner2 = new Tern(d3,0,0);
		Tern cg2 = Tern.transformation(m22, cgInner2).add(cb);
		Atom CG2 = new Atom("CG2",cg2);	
		
		ArrayList<Atom> list = new ArrayList<Atom>();
		list.add(CB);
		list.add(CG1);
		list.add(CG2);
		return list;
	}
	
	public static void main(String[] args) throws IOException
	{
		String file1 = "/home/pengx/1cy5.pdb";
		PDB pdb = new PDB(file1);
		Residue res = pdb.getFirstChain().getResidue(22);
		Tern cog = res.getExpectedCOG2();
		Tern cog1 = res.getExpectedCOG();
		Tern ca = res.getAtom("CA").getCoordinate();
		Tern cb = res.getAtom("CB").getCoordinate();
		
		double d1 = cog.distance(cog1);
		double d2 = cog.distance(ca);
		double d3 = cog.distance(cb);
		
		System.out.println(d1 + " " + d2 + " " + d3);
	}
}

