package model;
import java.io.*;
import similarSearch.*;
import java.util.*;

public class RotPair 
{
	private Residue resA;
	private Residue resB;
	
	private String pdbID="xxxx";
	private char chainID='-';
	
	private int seqDistance;
	private String rotA;
	private String rotB;
	
	ArrayList<Tern> resAFourAtoms;
	ArrayList<Tern> resBFourAtoms;
	
	ArrayList<Tern> eightAtoms;
	
	public StrideRes strResA;
	public StrideRes strResB;
	public boolean hasStrInfo=false;
	
	public RotPair(){};
	
	public RotPair(Residue resA, Residue resB)
	{
		this.resA = resA;
		this.resB = resB;
		
		if(!this.resA.infoSufficient() || !this.resB.infoSufficient())
		{
			throw new RuntimeException("residue info not sufficient");
		}
		
		this.seqDistance = Math.abs(resA.getID()-resB.getID());
		this.rotA = new SideChain(resA).getDunbrackCode();
		this.rotB = new SideChain(resB).getDunbrackCode();
		this.resAFourAtoms = this.resA.getBackboneFourAtomTerns();
		this.resBFourAtoms = this.resB.getBackboneFourAtomTerns();
		this.eightAtoms = new ArrayList<Tern>();
		for(int i=0;i<4;i++)
		{
			eightAtoms.add(this.resAFourAtoms.get(i));
		}
		for(int i=0;i<4;i++)
		{
			eightAtoms.add(this.resBFourAtoms.get(i));
		}
	}
	
	public RotPair(Residue resA, Residue resB, StrideRes srA, StrideRes srB, int seqDistance)
	{
		this.resA = resA;
		this.resB = resB;
		
		if(!this.resA.infoSufficient() || !this.resB.infoSufficient())
		{
			throw new RuntimeException("residue info not sufficient");
		}
		
		this.seqDistance = seqDistance;
		this.rotA = new SideChain(resA).getDunbrackCode();
		this.rotB = new SideChain(resB).getDunbrackCode();
		this.resAFourAtoms = this.resA.getBackboneFourAtomTerns();
		this.resBFourAtoms = this.resB.getBackboneFourAtomTerns();
		this.eightAtoms = new ArrayList<Tern>();
		for(int i=0;i<4;i++)
		{
			eightAtoms.add(this.resAFourAtoms.get(i));
		}
		for(int i=0;i<4;i++)
		{
			eightAtoms.add(this.resBFourAtoms.get(i));
		}
		this.strResA = srA;
		this.strResB = srB;
		this.hasStrInfo = true;
	}
	
	public RotPair(Residue resA, Residue resB, String pdbID, char chainID)
	{
		this(resA,resB);
		this.pdbID = pdbID;
		if(this.pdbID.length()!=4)
		{
			throw new RuntimeException("pdbID size:" + this.pdbID.length() + " != 4");
		}
		this.chainID = chainID;
	}
	
	public void addStrInfo(StrideProtein sp)
	{
		StrideRes sResA=sp.getStrideResidue(""+chainID+resA.getID());
		StrideRes sResB=sp.getStrideResidue(""+chainID+resB.getID());
		this.strResA=sResA;
		this.strResB=sResB;
		if(sResA!=null && sResB!=null)
		{
			this.hasStrInfo=true;
		}
		return;
	}
	
	public void setChainID(char chainID)
	{
		this.chainID = chainID;
	}
	
	public Residue getResA()
	{
		return this.resA;
	}
	
	public Residue getResB()
	{
		return this.resB;
	}
	
	public boolean hasStrideInfo()
	{
		return this.hasStrInfo;
	}
	
	public int seqDistance()
	{
		return this.seqDistance;
	}
	
	public double rmsd(RotPair other)
	{
		return math.RMSD.rmsd(this.eightAtoms, other.eightAtoms);
	}
	
	public double rmsd(String line)
	{
		if(line.length() < 201)
		{
			throw new RuntimeException("line length not enough: " + line.length());
		}
		
		ArrayList<Tern> ternList = new ArrayList<Tern>();
		for(int i=0;i<8;i++)
		{
			float x = Float.parseFloat(line.substring(33+i*21,33+i*21+7).trim());
			float y = Float.parseFloat(line.substring(33+i*21+7,33+i*21+14).trim());
			float z = Float.parseFloat(line.substring(33+i*21+14,33+i*21+21).trim());
			ternList.add(new Tern(x,y,z));
		}
		return math.RMSD.rmsd(this.eightAtoms, ternList);
	}
	
	public RotPair mirror()
	{
		RotPair mirror = new RotPair(this.resB,this.resA,this.pdbID,this.chainID);
		mirror.strResA = this.strResB;
		mirror.strResB = this.strResA;
		mirror.hasStrInfo = true;
		return mirror;
	}
	
	public String toString()
	{
		//order: pdbID ,chainID, resAIDSD, rotA, resBIDSD, rotB, resAFourAtomTern, resBFourAtomTern, resASS,resAPhi,resAPsi,resASasa,resBSS,resBPhi,resBPsi,resBSasa
		
		String pdbIDFormat=String.format("%4s", this.pdbID);
		String line=pdbIDFormat+" "+this.chainID+" ";
		line=line+String.format("%5d", this.resA.getIDSD());
		line=line+" "+this.rotA;
		line=line+String.format("%5d", this.resB.getIDSD());
		line=line+" "+this.rotB;
		for(int i=0;i<4;i++)
		{
			line=line + String.format("%7.2f", this.resAFourAtoms.get(i).x());
			line=line + String.format("%7.2f", this.resAFourAtoms.get(i).y());
			line=line + String.format("%7.2f", this.resAFourAtoms.get(i).z());
		}
		for(int i=0;i<4;i++)
		{
			line=line + String.format("%7.2f", this.resBFourAtoms.get(i).x());
			line=line + String.format("%7.2f", this.resBFourAtoms.get(i).y());
			line=line + String.format("%7.2f", this.resBFourAtoms.get(i).z());
		}
		if(this.hasStrInfo)
		{
			line=line+this.strResA.toString()+this.strResB.toString();
		}
		return line;
	}
	
	public static void main(String[] args) throws IOException
	{
		PDB pdb = new PDB("E:\\pairwiseDesign\\testPDB2\\mod2\\native.pdb");
		StrideProtein sp = new StrideProtein("E:\\pairwiseDesign\\testPDB2\\mod2\\mod2_mea.str");
		ArrayList<RotPair> rps = pdb.getChains().get(0).findContactRotPairs(8.0);
		PrintStream out = new PrintStream("E:\\pairwiseDesign\\pii.rotPairs.txt");
		for(RotPair rp : rps)
		{
			rp.addStrInfo(sp);
			out.println(rp.toString());
		}
		out.close();
		
		RotPair rp = rps.get(0);
		for(RotPair rp2 : rps)
		{
			String line = rp2.toString();
			System.out.println(rp.rmsd(line));
		}
	}
}
