package similarSearch;

import java.util.ArrayList;

import math.RMSD;
import model.Tern;
import model.ResPair;

public class PairRelativePosition 
{
	private double caCaDistance;
	private double angleA;
	private double angleB;
	private double dihedral;
	
	public PairRelativePosition(double caCaDistance,double angleA,double angleB,double dihedral)
	{
		this.caCaDistance=caCaDistance;
		this.angleA=angleA;
		this.angleB=angleB;
		this.dihedral=dihedral;
	}
	
	public ArrayList<Tern> generateCaAlignedTerns(double bondLength)
	{
		
		Tern A=new Tern(-0.5*this.caCaDistance,0,0);
		Tern B=new Tern(0.5*this.caCaDistance,0,0);
		double xC=bondLength*Math.cos(Math.toRadians(this.angleA))-0.5*this.caCaDistance;
		double yC=bondLength*Math.sin(Math.toRadians(this.angleA))*Math.cos(Math.toRadians(dihedral/2));
		double zC=bondLength*Math.sin(Math.toRadians(this.angleA))*Math.sin(Math.toRadians(dihedral/2));
		Tern C=new Tern(xC,yC,zC);
		double xD=-1*bondLength*Math.cos(Math.toRadians(this.angleB))+0.5*this.caCaDistance;
		double yD=bondLength*Math.sin(Math.toRadians(this.angleB))*Math.cos(Math.toRadians(dihedral/2));
		double zD=-1*bondLength*Math.sin(Math.toRadians(this.angleB))*Math.sin(Math.toRadians(dihedral/2));
		Tern D=new Tern(xD,yD,zD);
		ArrayList<Tern> ternList=new ArrayList<Tern>();
		ternList.add(A);
		ternList.add(B);
		ternList.add(C);
		ternList.add(D);
		return ternList;
	}
	
	public ArrayList<Tern> generateCaAlignedCaCbTerns()
	{
		//we set the CaCb distance 1.54		
		return generateCaAlignedTerns(1.54);
	}
	
	public ArrayList<Tern> generateCaAlignedCaEcogTerns()
	{
		//we set the distance between Ca atom and the center of gravity of sidechain 2.549 
		return generateCaAlignedTerns(2.549);
	}
	
	public ArrayList<Tern> generateEcogAlignedCaEcogTerns()
	{
		ArrayList<Tern> caAlignedTerns=this.generateCaAlignedCaEcogTerns();
		double ecogDistance=caAlignedTerns.get(2).distance(caAlignedTerns.get(3));
		Tern GAGB=caAlignedTerns.get(3).subtract(caAlignedTerns.get(2));
		Tern GBGA=GAGB.negate();
		Tern GACA=caAlignedTerns.get(0).subtract(caAlignedTerns.get(2));
		Tern GBCB=caAlignedTerns.get(1).subtract(caAlignedTerns.get(3));
		
		double angGA=GAGB.angle(GACA);
		double angGB=GBGA.angle(GBCB);

		Tern n1=GAGB.cross(GACA.cross(GAGB));
		Tern n2=GBGA.cross(GBCB.cross(GBGA));
		double dihedralG=n1.angle(n2);
		if(dihedralG!=0)
		{
			double symbol=n1.cross(n2).dotProduct(GBGA);
			symbol=symbol/Math.abs(symbol);
			dihedralG=dihedralG*symbol;
		}
		
		PairRelativePosition newPrp=new PairRelativePosition(ecogDistance,angGA,angGB,dihedralG);
		
		ArrayList<Tern> GFirstTerns=newPrp.generateCaAlignedCaEcogTerns();
		
		ArrayList<Tern> CFirstTerns=new ArrayList<Tern>();
		CFirstTerns.add(GFirstTerns.get(2));
		CFirstTerns.add(GFirstTerns.get(3));
		CFirstTerns.add(GFirstTerns.get(0));
		CFirstTerns.add(GFirstTerns.get(1));
		return CFirstTerns;
	}
	
	public double rmsd1(PairRelativePosition other)
	{
		ArrayList<Tern> list1=this.generateCaAlignedCaCbTerns();
		ArrayList<Tern> list2=other.generateCaAlignedCaCbTerns();
		return RMSD.simpleRmsd(list1, list2);
	}
	
	public double maxDistance(PairRelativePosition other)
	{
		ArrayList<Tern> list1 = this.generateCaAlignedCaCbTerns();
		ArrayList<Tern> list2 = other.generateCaAlignedCaCbTerns();
		double maxDis = 0.0;
		for(int i=0; i<list1.size();i++)
		{
			double dis = list1.get(0).distance(list2.get(0));
			if(dis > maxDis)
			{
				maxDis = dis;
			}
		}
		return maxDis;
	}
	
	public double rmsd2(PairRelativePosition other,double lamda)
	{
		double rmsdSquare=0.0;
		ArrayList<Tern> list1=this.generateEcogAlignedCaEcogTerns();
		ArrayList<Tern> list2=other.generateEcogAlignedCaEcogTerns();
		rmsdSquare+=1.0/(1.0+lamda)*list1.get(0).distSquare(list2.get(0));
		rmsdSquare+=1.0/(1.0+lamda)*list1.get(1).distSquare(list2.get(1));
		rmsdSquare+=lamda/(1.0+lamda)*list1.get(2).distSquare(list2.get(2));
		rmsdSquare+=lamda/(1.0+lamda)*list1.get(3).distSquare(list2.get(3));
		return Math.sqrt(rmsdSquare);
	}
	
	public double rmsd3(PairRelativePosition other)
	{
		ArrayList<Tern> list1=this.generateCaAlignedCaEcogTerns();
		ArrayList<Tern> list2=other.generateCaAlignedCaEcogTerns();
		return RMSD.simpleRmsd(list1, list2);
	}
	
	public static PairRelativePosition center(PairRelativePosition prp1, double weight1, PairRelativePosition prp2, double weight2)
	{
		double caCaDistance = (prp1.caCaDistance * weight1 + prp2.caCaDistance*weight2)/(weight1+weight2);
		double angleA = (prp1.angleA * weight1 + prp2.angleA*weight2)/(weight1+weight2);
		double angleB= (prp1.angleB * weight1 + prp2.angleB*weight2)/(weight1+weight2);
		double dihedral = (prp1.dihedral * weight1 + prp2.dihedral*weight2)/(weight1+weight2);
		return new PairRelativePosition(caCaDistance,angleA,angleB,dihedral);
	}
}
