package similarSearch;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import math.ProbabilityArrayRT;
import math.ProbabilityMatrix;
import math.ProbabilityMatrixRT;
import model.InterfacePair;
import model.InterfaceResidue;
import model.ResName;
import model.Rotamers;

public class PiPairDatabase implements Serializable{

	
	/**
	 * 
	 */
	private static final long serialVersionUID = -1006680355868722255L;
	HashMap<String, ArrayList<InterfacePair>> interPairs = new HashMap<String, ArrayList<InterfacePair>>();
	HashMap<String, ArrayList<InterfacePair>> intraPairs = new HashMap<String, ArrayList<InterfacePair>>();
	
	public PiPairDatabase(String interFile, String intraFile) throws IOException
	{
		BufferedReader in = new BufferedReader(new FileReader(interFile));
		String line;
		while((line = in.readLine())!=null)
		{
			if(line.length() < 303) continue;
			InterfacePair ip = new InterfacePair(line);
			String key = getKey(ip);
			if(!this.interPairs.containsKey(key))
				this.interPairs.put(key, new ArrayList<InterfacePair>());
			this.interPairs.get(key).add(ip);
		}
		in.close();
		
		in = new BufferedReader(new FileReader(intraFile));
		while((line = in.readLine())!=null)
		{
			if(line.length() < 303) continue;
			InterfacePair ip = new InterfacePair(line);
			String key = getKey(ip);
			if(!this.intraPairs.containsKey(key))
				this.intraPairs.put(key, new ArrayList<InterfacePair>());
			this.intraPairs.get(key).add(ip);
		}
		in.close();
	}
	
	public static String getKey(InterfacePair ip)
	{
		if(ip.sameChain)
		{
			int seqDist = Math.abs(ip.irA.getResidue().getID() - ip.irB.getResidue().getID());
			if(seqDist > 5) seqDist = 5;
			
			String key = seqDist + "-" + ip.irA.ss + ip.irB.ss + "-" + PiResDatabase.saiIndex(ip.irA.saiD) + "-" + PiResDatabase.saiIndex(ip.irB.saiD);
			return key;
		}
		else
		{
			String key = ""+ ip.irA.ss + ip.irB.ss + "-" + PiResDatabase.saiIndex(ip.irA.saiD) + "-" + PiResDatabase.saiIndex(ip.irB.saiD);
			return key;
		}
	}
	
	
	public ProbabilityMatrixRT findSimilarPairsRT(InterfacePair ipQuery, PairPara ppr, Rotamers rots, HashSet<String> banPDBs)
	{
		if(ipQuery.sameChain)
			return findSimilarIntraPairsRT(ipQuery,ppr,rots,banPDBs);
		else
			return findSimilarInterPairsRT(ipQuery, ppr,rots,banPDBs);
	}
	
	public ProbabilityMatrixRT findSimilarIntraPairsRT(InterfacePair ipQuery, PairPara ppr, Rotamers rots, HashSet<String> banPDBs)
	{
		InterfacePair ipQueryMirror = ipQuery.mirror();
		HashSet<String> selectKeySet = new HashSet<String>();
		ResPara rpr = ppr.resParaA();
		for(String key : this.intraPairs.keySet())
		{	
			int seqDist = Integer.parseInt(key.substring(0, 1));
			if(seqDist != ipQuery.seqDistance()) continue;
			
			char ssA = key.charAt(2);
			boolean ssAFit = ResCharacter.simSecondaryStructure(ssA, ipQuery.irA.ss, rpr.ssLevel());
			if(!ssAFit)
				continue;
			
			char ssB = key.charAt(3);
			boolean ssBFit = ResCharacter.simSecondaryStructure(ssB, ipQuery.irB.ss, rpr.ssLevel());
			if(!ssBFit)
				continue;
			
			
			double saiDALow = 0.1*Integer.parseInt(key.substring(5, 6));
			double saiDAHigh = saiDALow + 0.1;
			boolean saiDAFit = ResCharacter.simSasaN(ipQuery.irA.saiD, saiDALow,rpr.sasaLevel()) || ResCharacter.simSasaN(ipQuery.irA.saiD, saiDAHigh,rpr.sasaLevel());
			if(ipQuery.irA.saiD > saiDALow && ipQuery.irA.saiD < saiDAHigh)
				saiDAFit = true;
			if(!saiDAFit)
				continue;
			
			
			double saiDBLow = 0.1*Integer.parseInt(key.substring(7, 8));
			double saiDBHigh = saiDBLow + 0.1;
			boolean saiDBFit = ResCharacter.simSasaN(ipQuery.irB.saiD, saiDBLow,rpr.sasaLevel()) || ResCharacter.simSasaN(ipQuery.irB.saiD, saiDBHigh,rpr.sasaLevel());
			if(ipQuery.irB.saiD > saiDBLow && ipQuery.irB.saiD < saiDBHigh)
				saiDBFit = true;
			if(!saiDBFit)
				continue;
			
			selectKeySet.add(key);		
		}		
		
		double[][] pm = new double[rots.rotamerNum()][rots.rotamerNum()];
		int count = 0;
		for(String key : selectKeySet)
		{
			for(InterfacePair ip : this.intraPairs.get(key))
			{
				if(banPDBs.contains(ip.irA.pdbID))
					continue;
				if(ip.similarTo(ipQuery, ppr))
				{
					int rotAID = rots.dunbrackCodeToInt(ip.irA.getDunbrackCode());
					int rotBID = rots.dunbrackCodeToInt(ip.irB.getDunbrackCode());
					pm[rotAID][rotBID] += 1.0;
					count++;
				}
				if(ip.similarTo(ipQueryMirror, ppr))
				{
					int rotAID = rots.dunbrackCodeToInt(ip.irA.getDunbrackCode());
					int rotBID = rots.dunbrackCodeToInt(ip.irB.getDunbrackCode());
					pm[rotBID][rotAID] += 1.0;
					count++;
				}
			}
		}
		return new ProbabilityMatrixRT(pm, count);
	}
	
	public ProbabilityMatrixRT findSimilarInterPairsRT(InterfacePair ipQuery, PairPara ppr,Rotamers rots, HashSet<String> banPDBs)
	{
		InterfacePair ipQueryMirror = ipQuery.mirror();
		HashSet<String> selectKeySet = new HashSet<String>();
		ResPara rpr = ppr.resParaA();
		for(String key : this.interPairs.keySet())
		{
			int seqDist = Integer.parseInt(key.substring(0, 1));
			if(seqDist != ipQuery.seqDistance()) continue;
			
			char ssA = key.charAt(2);
			boolean ssAFit = ResCharacter.simSecondaryStructure(ssA, ipQuery.irA.ss, rpr.ssLevel());
			if(!ssAFit)
				continue;
			
			char ssB = key.charAt(3);
			boolean ssBFit = ResCharacter.simSecondaryStructure(ssB, ipQuery.irB.ss, rpr.ssLevel());
			if(!ssBFit)
				continue;
			
			
			double saiDALow = 0.1*Integer.parseInt(key.substring(5, 6));
			double saiDAHigh = saiDALow + 0.1;
			boolean saiDAFit = ResCharacter.simSasaN(ipQuery.irA.saiD, saiDALow,rpr.sasaLevel()) || ResCharacter.simSasaN(ipQuery.irA.saiD, saiDAHigh,rpr.sasaLevel());
			if(ipQuery.irA.saiD > saiDALow && ipQuery.irA.saiD < saiDAHigh)
				saiDAFit = true;
			if(!saiDAFit)
				continue;
			
			
			double saiDBLow = 0.1*Integer.parseInt(key.substring(7, 8));
			double saiDBHigh = saiDBLow + 0.1;
			boolean saiDBFit = ResCharacter.simSasaN(ipQuery.irB.saiD, saiDBLow,rpr.sasaLevel()) || ResCharacter.simSasaN(ipQuery.irB.saiD, saiDBHigh,rpr.sasaLevel());
			if(ipQuery.irB.saiD > saiDBLow && ipQuery.irB.saiD < saiDBHigh)
				saiDBFit = true;
			if(!saiDBFit)
				continue;
			
			selectKeySet.add(key);	
		}
		
		double[][] pm = new double[rots.rotamerNum()][rots.rotamerNum()];
		int count = 0;
		for(String key : selectKeySet)
		{
			for(InterfacePair ip : this.interPairs.get(key))
			{
				if(banPDBs.contains(ip.irA.pdbID))
					continue;
				if(ip.similarTo(ipQuery, ppr))
				{
					int rotAID = rots.dunbrackCodeToInt(ip.irA.getDunbrackCode());
					int rotBID = rots.dunbrackCodeToInt(ip.irB.getDunbrackCode());
					pm[rotAID][rotBID] += 1.0;
					count++;
				}
				if(ip.similarTo(ipQueryMirror, ppr))
				{
					int rotAID = rots.dunbrackCodeToInt(ip.irA.getDunbrackCode());
					int rotBID = rots.dunbrackCodeToInt(ip.irB.getDunbrackCode());
					pm[rotBID][rotAID] += 1.0;
					count++;
				}
			}
		}
		return new ProbabilityMatrixRT(pm, count);
	}
	
	public ProbabilityMatrix findSimilarPairs(InterfacePair ipQuery, PairPara ppr, HashSet<String> banPDBs)
	{
		if(ipQuery.sameChain)
			return findSimilarIntraPairs(ipQuery,ppr,banPDBs);
		else
			return findSimilarInterPairs(ipQuery, ppr,banPDBs);
	}
	
	public ProbabilityMatrix findSimilarIntraPairs(InterfacePair ipQuery, PairPara ppr,  HashSet<String> banPDBs)
	{
		InterfacePair ipQueryMirror = ipQuery.mirror();
		HashSet<String> selectKeySet = new HashSet<String>();
		
		for(String key : this.intraPairs.keySet())
		{
			if(key.length() != 8) continue;
			int seqDist = Integer.parseInt(key.substring(0, 1));
			if(seqDist != ipQuery.seqDistance()) continue;
			char ssA = key.charAt(2);
			boolean ssFitA = ResCharacter.simSecondaryStructure(ssA, ipQuery.irA.ss, ppr.resParaA().ssLevel());
			char ssB = key.charAt(3);
			boolean ssFitB = ResCharacter.simSecondaryStructure(ssB, ipQuery.irB.ss, ppr.resParaA().ssLevel());
			if(!ssFitA || !ssFitB) continue;
			
			
			double saiDALow = 0.1*Integer.parseInt(key.substring(5, 6));
			double saiDAHigh = saiDALow + 0.1;
			
			
			double saiDBLow = 0.1*Integer.parseInt(key.substring(7, 8));
			double saiDBHigh = saiDBLow + 0.1;
			
			boolean saFitA = ResCharacter.simSasaN(ipQuery.irA.saiD, saiDALow,ppr.resParaA().sasaLevel()) || ResCharacter.simSasaN(ipQuery.irA.saiD, saiDAHigh,ppr.resParaA().sasaLevel());
			boolean saFitB = ResCharacter.simSasaN(ipQuery.irB.saiD, saiDBLow,ppr.resParaA().sasaLevel()) || ResCharacter.simSasaN(ipQuery.irB.saiD, saiDBHigh,ppr.resParaA().sasaLevel());
			if(!saFitA || !saFitB) continue;
			//System.out.println(key);
			
			selectKeySet.add(key);		
		}		
		
		double[][] pm = new double[20][20];
		int count = 0;
		for(String key : selectKeySet)
		{
			for(InterfacePair ip : this.intraPairs.get(key))
			{
				if(banPDBs.contains(ip.irA.pdbID))
					continue;
				if(ip.similarTo(ipQuery, ppr))
				{
					int resAType = ResName.triToInt(ip.irA.getResType());
					int resBType = ResName.triToInt(ip.irB.getResType());
					pm[resAType][resBType] += 1.0;
					count++;
				}
				if(ip.similarTo(ipQueryMirror, ppr))
				{
					int resAType = ResName.triToInt(ip.irA.getResType());
					int resBType = ResName.triToInt(ip.irB.getResType());
					pm[resBType][resAType] += 1.0;
					count++;
				}
			}
		}
		return new ProbabilityMatrix(pm, count);
	}
	
	public ProbabilityMatrix findSimilarInterPairs(InterfacePair ipQuery, PairPara ppr, HashSet<String> banPDBs)
	{
		InterfacePair ipQueryMirror = ipQuery.mirror();
		HashSet<String> selectKeySet = new HashSet<String>();
		for(String key : this.interPairs.keySet())
		{
			//System.out.println("x: " + key + " " + this.interPairs.get(key).size());
			//InterfacePair ip = this.interPairs.get(key).get(0);
			//System.out.println(ip.sameChain);
			//System.out.println(getKey(ip));
			//char a = this.interPairs.get(key).get(0).irA.chainID;
			//char b = this.interPairs.get(key).get(0).irB.chainID;
			//System.out.println(a + " " + b);
			if(key.length() != 6) continue;
			char ssA = key.charAt(0);
			boolean ssFitA = ResCharacter.simSecondaryStructure(ssA, ipQuery.irA.ss, ppr.resParaA().ssLevel());
		
			char ssB = key.charAt(1);
			boolean ssFitB = ResCharacter.simSecondaryStructure(ssB, ipQuery.irB.ss, ppr.resParaA().ssLevel());
			if(!ssFitA || !ssFitB) continue;
			
			double saiDALow = 0.1*Integer.parseInt(key.substring(3, 4));
			double saiDAHigh = saiDALow + 0.1;
			
			boolean saFitA = ResCharacter.simSasaN(ipQuery.irA.saiD, saiDALow,ppr.resParaA().sasaLevel()) || ResCharacter.simSasaN(ipQuery.irA.saiD, saiDAHigh,ppr.resParaA().sasaLevel());
			
			
			double saiDBLow = 0.1*Integer.parseInt(key.substring(5, 6));
			double saiDBHigh = saiDBLow + 0.1;
			
			boolean saFitB = ResCharacter.simSasaN(ipQuery.irB.saiD, saiDBLow,ppr.resParaA().sasaLevel()) || ResCharacter.simSasaN(ipQuery.irB.saiD, saiDBHigh,ppr.resParaA().sasaLevel());
			
			if(!saFitA || !saFitB) continue;
			//System.out.println(key);
			selectKeySet.add(key);
		}
		
		double[][] pm = new double[20][20];
		int count = 0;
		for(String key : selectKeySet)
		{
			for(InterfacePair ip : this.interPairs.get(key))
			{
				if(banPDBs.contains(ip.irA.pdbID))
					continue;
				if(ip.similarTo(ipQuery, ppr))
				{
					int resAType = ResName.triToInt(ip.irA.getResType());
					int resBType = ResName.triToInt(ip.irB.getResType());
					pm[resAType][resBType] += 1.0;
					count++;
				}
				if(ip.similarTo(ipQueryMirror, ppr))
				{
					int resAType = ResName.triToInt(ip.irA.getResType());
					int resBType = ResName.triToInt(ip.irB.getResType());
					pm[resBType][resAType] += 1.0;
					count++;
				}
			}
		}
		return new ProbabilityMatrix(pm, count);
	}
	
	public void saveToDisk(File file) throws IOException
    {
        ZipOutputStream zout = new ZipOutputStream(new FileOutputStream(file));
        ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(zout));
        ZipEntry ze = new ZipEntry("piPairDatabase");
        zout.putNextEntry(ze);
        out.writeObject(this);
        out.flush();
        out.close();
        zout.close();
    }
    
	public static PiPairDatabase loadFromDisk(File file) throws IOException
	{
		ZipInputStream zin = new ZipInputStream(new  FileInputStream(file));
		ZipEntry ze = zin.getNextEntry();
		if(!ze.getName().equals("piPairDatabase"))
		{
			zin.close();
			throw new RuntimeException("Error: not a piPairDatabase file");
		}
		ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(zin));
		PiPairDatabase piPairDatabase = null;
		try{
			piPairDatabase = (PiPairDatabase) in.readObject();
		}
		catch(ClassNotFoundException exception)
		{
			System.err.println("Error: not a piPairDatabase file");
			exception.printStackTrace(System.err);
			System.exit(1);
		}
		in.close();
		zin.close();
		return piPairDatabase;
	}
	
	public static void main(String[] args) throws IOException
	{
		 String fileInter = "/nfs/amino-home/pengx/ppi/piface/piPairs/interPair.represent-8.0.txt";
		 
		 String fileIntra = "/nfs/amino-home/pengx/ppi/piface/piPairs/intraPair.represent-8.0.txt";
		 File file = new File("/nfs/amino-home/pengx/ppi/piface/piPairs/piPair.represent-8.0.dat");
		 PiPairDatabase pairDatabase = new PiPairDatabase(fileInter, fileIntra); 
		 pairDatabase.saveToDisk(file);
	}
}
