package similarSearch;
import java.io.*;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import math.ProbabilityArray;
import math.ProbabilityArrayRT;
import model.InterfacePair;
import model.InterfaceResidue;
import model.ResName;
import model.Rotamers;


public class PiResDatabase implements Serializable{

	
	private static final long serialVersionUID = 3667001859433369585L;
	public final HashMap<String,ArrayList<InterfaceResidue>> piResidues = new HashMap<String, ArrayList<InterfaceResidue>>();

	 
	 public PiResDatabase(String piResFile) throws IOException
	 {
		 BufferedReader in = new BufferedReader(new FileReader(piResFile));
		 String line;
		 while((line = in.readLine())!=null)
		 {
			 InterfaceResidue ir = new InterfaceResidue(line);
			 String key = ir.ss + "-" +saiIndex(ir.saiD);
			 if(!piResidues.containsKey(key))
			 {
				 piResidues.put(key, new ArrayList<InterfaceResidue>());
			 }
			 piResidues.get(key).add(ir);
		 }
		 in.close();
	 }
	 
	
	 
	 public void saveToDisk(File file) throws IOException
	    {
	        ZipOutputStream zout = new ZipOutputStream(
	                new FileOutputStream(file));
	        ObjectOutputStream out = new ObjectOutputStream(
	                new BufferedOutputStream(zout));
	        ZipEntry ze = new ZipEntry("piResDatabase");
	        zout.putNextEntry(ze);
	        out.writeObject(this);
	        out.flush();
	        out.close();
	        zout.close();
	    }
	    
	 public static PiResDatabase loadFromDisk(File file) throws IOException
     {
        ZipInputStream zin = new ZipInputStream(
                new  FileInputStream(file));
        ZipEntry ze = zin.getNextEntry();
        if(!ze.getName().equals("piResDatabase"))
        {
        	zin.close();
        	throw new RuntimeException("Error: not a piResDatabase file");
        }
        ObjectInputStream in = new ObjectInputStream(
                new BufferedInputStream(zin));
        PiResDatabase piResDatabase = null;
        try{
        	piResDatabase = (PiResDatabase) in.readObject();
        }
        catch(ClassNotFoundException exception)
        {
            System.err.println("Error: not a piResDatabase file");
            exception.printStackTrace(System.err);
            System.exit(1);
        }
        in.close();
        zin.close();
        return piResDatabase;
     }
	 
	 public ProbabilityArrayRT findSimilarResRT(ResPara rpr, InterfaceResidue irQuery, Rotamers rots, HashSet<String> banPDBs)
	{
		HashSet<String> selectKeySet = new HashSet<String>();
		for(String key : this.piResidues.keySet())
		{
			char ssType = key.charAt(0);
			double saiDLow = 0.1*Integer.parseInt(key.substring(2, 3));
			double saiDHigh = saiDLow + 0.1;
			boolean ssFit = ResCharacter.simSecondaryStructure(ssType, irQuery.ss, rpr.ssLevel());
			if(!ssFit)
				continue;
			boolean saiDFit = ResCharacter.simSasaN(irQuery.saiD, saiDLow,rpr.sasaLevel()) || ResCharacter.simSasaN(irQuery.saiD, saiDHigh,rpr.sasaLevel());
			if(irQuery.saiD > saiDLow && irQuery.saiD < saiDHigh)
				saiDFit = true;
			if(!saiDFit)
				continue;
			selectKeySet.add(key);
		}
		
		double[] p = new double[rots.rotamerNum()];
		int count = 0;
		
		for(String key : selectKeySet)
		{
		
			for(InterfaceResidue ir : this.piResidues.get(key))
			{
				if(banPDBs.contains(ir.pdbID)) 
					continue;
				if(ir.similarTo(irQuery, rpr))
				{
					p[rots.dunbrackCodeToInt(ir.dunbrackCode)] += 1.0;
					count++;
				}
				
			}
		}
		return new ProbabilityArrayRT(p,count);
	}
	 
	 public ProbabilityArray findSimilarRes(ResPara rpr, InterfaceResidue irQuery, HashSet<String> banPDBs)
	{
			HashSet<String> selectKeySet = new HashSet<String>();
			for(String key : this.piResidues.keySet())
			{
				char ssType = key.charAt(0);
				double saiDLow = 0.1*Integer.parseInt(key.substring(2, 3));
				double saiDHigh = saiDLow + 0.1;
				boolean ssFit = ResCharacter.simSecondaryStructure(ssType, irQuery.ss, rpr.ssLevel());
				if(!ssFit)
					continue;
				boolean saiDFit = ResCharacter.simSasaN(irQuery.saiD, saiDLow,rpr.sasaLevel()) || ResCharacter.simSasaN(irQuery.saiD, saiDHigh,rpr.sasaLevel());
				if(irQuery.saiD > saiDLow && irQuery.saiD < saiDHigh)
					saiDFit = true;
				if(!saiDFit)
					continue;
				selectKeySet.add(key);
			}
			
			double[] p = new double[20];
			int count = 0;
			
			for(String key : selectKeySet)
			{
			
				for(InterfaceResidue ir : this.piResidues.get(key))
				{
					if(banPDBs.contains(ir.pdbID)) 
						continue;
					if(ir.similarTo(irQuery, rpr))
					{
						int type = ResName.triToInt(ir.getResType());
						p[type] += 1.0;
						count++;
					}
					
				}
			}
			return new ProbabilityArray(p,count);
	}
	 
	 public static int saiIndex(double sai)
	 {
		 int index = (int)Math.floor(sai*10);
		 if(index == 10) return 9;
		 else return index;
	 }
	 
	 public static void main(String[] args) throws IOException
	 {
		 String s = "/nfs/amino-home/pengx/ppi/piface/piResidues/piResidues.represent-8.0.txt";
		 File file = new File("/nfs/amino-home/pengx/ppi/piface/piResidues/piRes.represent-8.0.dat");
		 PiResDatabase piResDatabase = new PiResDatabase(s);
		 piResDatabase.saveToDisk(file);
	 }
}
