package thermo;
import java.io.*;
import java.util.*;

import similarSearch.FindSimilarPairsRT;
import similarSearch.FindSimilarResiduesRT;
import similarSearch.PairPara;
import similarSearch.ResPara;
import math.ProbabilityArrayRT;
import math.ProbabilityMatrixRT;
import math.ScoreMatrixRT;
import model.*;
import input.InputFile;

public class SingleMutationScanning 
{
	
	public static ArrayList<Double> getS1List(PDB pdb, StrideProtein sp, int resID, ResPara rpr, Rotamers rots) throws IOException
	{
		int resIDSD = pdb.getFirstChain().getResidue(resID).getIDSD();
		
		StrideRes sr = sp.getStrideChains().get(0).getResidues().get(resIDSD);
		HashSet<String> banList = new HashSet<String>();
		ProbabilityArrayRT pa = similarSearch.FindSimilarResiduesRT.findPAofRes(sr, rpr, rots, banList);
		ProbabilityArrayRT Pb=ProbabilityArrayRT.getPB(4);
		pa.addPseudoCount(Pb);
		ProbabilityArrayRT pax = SingleSiteScoring.toAADistributionOverBackground(pa);
		
		
		ArrayList<Integer> mutRotTypeList = new ArrayList<Integer>();
		for(int i=0;i<20;i++)
		{
			mutRotTypeList.add(rots.getAminoAcidRotamerID(i).get(0));
		}
		ArrayList<Double> s1List = new ArrayList<Double>();
		for(int i=0;i<20;i++)
		{
			double p = pax.getArray().get(mutRotTypeList.get(i));
			s1List.add(-1.0*Math.log(p));
		}
		return s1List;
	}
	
	public static ArrayList<Double> getS2List(PDB pdb, StrideProtein sp, int resID, PairPara ppr, Rotamers rots) throws IOException
	{
		int resIDSD = pdb.getFirstChain().getResidue(resID).getIDSD();
		ArrayList<RotPair> rps = pdb.getFirstChain().findContactRotPairs(8.0);
		ArrayList<RotPair> involvedPairs = new ArrayList<RotPair>();
		for(RotPair rp : rps)
		{
			if(rp.getResA().getIDSD() == resIDSD || rp.getResB().getIDSD() == resIDSD)
				involvedPairs.add(rp);
		}
		ArrayList<ScoreMatrixRT> smList = new ArrayList<ScoreMatrixRT>();
		ProbabilityArrayRT Pb=ProbabilityArrayRT.getPB(4);
		HashSet<String> banList = new HashSet<String>();
		
		for(RotPair rp : involvedPairs)
		{
			rp.addStrInfo(sp);
			ProbabilityMatrixRT pm = FindSimilarPairsRT.findPMofRotPair(rp, ppr, banList, rots);
			
			PairPara thisPpr = new PairPara(ppr.resParaA(),ppr.resParaB(),ppr.cutoff(),ppr.fqFloor());

			while(pm.maxFq()<thisPpr.fqFloor() && thisPpr.cutoff()<1.5)
			{
				thisPpr = thisPpr.demote();
				pm = FindSimilarPairsRT.findPMofRotPair(rp,thisPpr,banList,rots);
			}

			ProbabilityArrayRT paResA = FindSimilarResiduesRT.findPAofRes(rp.strResA, thisPpr.resParaA(),rots,banList);
			paResA.addPseudoCount(Pb);
			ProbabilityArrayRT paResB = FindSimilarResiduesRT.findPAofRes(rp.strResB, thisPpr.resParaB(),rots,banList);
			paResB.addPseudoCount(Pb);
			ProbabilityMatrixRT pmCross = paResA.cross(paResB);
			
			ScoreMatrixRT sm = new ScoreMatrixRT(pm,pmCross);
			
			sm.adjustment(pm);
			
			int seqDist = rp.seqDistance();
			if(seqDist < 3)
				sm.multiply(0.5);
			else if(seqDist < 5)
				sm.multiply(0.6);
			smList.add(sm);
		}
		
		ArrayList<Double> s2List = new ArrayList<Double>();
		String natRotSeq = pdb.getChains().get(0).getRotamerSeq();
		for(int i=0;i<20;i++)
		{
			char type = ResName.intToSin(i);
			ArrayList<String> mutSeqs = SingleSiteScoring.mutate(natRotSeq, resIDSD, type);
			ArrayList<Double> mutS2s = new ArrayList<Double>();
			for(String mutSeq : mutSeqs)
			{
				mutS2s.add(SingleSiteScoring.getS2(mutSeq,smList,involvedPairs,rots));
			}
			double minMutS2 = SingleSiteScoring.min(mutS2s);
			s2List.add(minMutS2);
		}
		
		return s2List;
	}
	
	
	public static void main(String[] args) throws IOException
	{
		
		String pdbFile = args[0];
		String mut = args[1];
		String workPath = args[2];
		
		File pdbF = new File(pdbFile);
		if(!pdbF.exists())
			throw new RuntimeException("can't file pdb file: " + pdbFile);
		
		String[] spt = pdbFile.split("/");
		String fileName = spt[spt.length-1];
		
		String pdbTitle = fileName.substring(0, fileName.length()-4);
		
		String psdFile = workPath + "/" + pdbTitle + ".psd";
		String strFile = workPath + "/" + pdbTitle + ".str";
		
	
		usr.BuildPseudoSidechain.rewriteToPSD(pdbFile, psdFile);
		String[] envp = new String[2];
		envp[0] = psdFile;
		envp[1] = ">" + strFile;
		
		InputFile input = new InputFile();
		String stride = input.get("stride");
		
		Process p = Runtime.getRuntime().exec(stride + " " + psdFile);
		BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(p.getInputStream()));
		String s;
		PrintStream out = new PrintStream(strFile);
		while((s=bufferedReader.readLine()) != null)
		{
			out.println(s);
		}
		out.close();
		
		
		File strF = new File(strFile);
		if(!strF.exists())
			throw new RuntimeException("fail to generate stride file");
		
		PDB pdb = new PDB(pdbFile);
		StrideProtein sp = new StrideProtein(strFile);
		
		int resID = Integer.parseInt(mut.substring(1, mut.length()-1));
		
		ResPara rpr = new ResPara(1, 4, 0);
		PairPara ppr = new PairPara(40, 0.5, 1, 2, 0);
		Rotamers rots = new Rotamers(4);
		ArrayList<Double> s1List = getS1List(pdb, sp, resID, rpr, rots);
		ArrayList<Double> s2List = getS2List(pdb, sp, resID, ppr, rots);
		
		int natType = ResName.triToInt(pdb.getFirstChain().getResidue(resID).getName());
		for(int i=0;i<20;i++)
		{
			double delS1 = s1List.get(i) - s1List.get(natType);
			double delS2 = s2List.get(i) - s2List.get(natType);
			double s1s2 = delS1 + delS2;
			System.out.printf("%s %6.3f\n",ResName.intToTri(i), s1s2);
		}
	}
}
