package thermo;
import java.io.*;
import java.util.*;

import similarSearch.FindSimilarPairsRT;
import similarSearch.FindSimilarResiduesRT;
import similarSearch.PairPara;
import similarSearch.ResPara;
import math.ProbabilityArrayRT;
import math.ProbabilityMatrixRT;
import math.ScoreMatrixRT;
import model.*;


public class SingleSiteScoring {
	
	public static ProbabilityArrayRT toAADistribution(ProbabilityArrayRT paOld) throws IOException
	{
		Rotamers rots = new Rotamers(4);
		double[] paN = new double[rots.rotamerNum()];
		for(int i=0;i<20;i++)
		{
			ArrayList<Integer> rotIDList = rots.getAminoAcidRotamerID(i);
			double p = 0;
			for(Integer rotID : rotIDList)
			{
				p = p + paOld.getArray().get(rotID);
			}
			for(Integer rotID : rotIDList)
			{
				paN[rotID] = p;
			}
		}
		return new ProbabilityArrayRT(paN);
	}
	
	public static ProbabilityArrayRT toAADistributionOverBackground(ProbabilityArrayRT paOld) throws IOException
	{
		ProbabilityArrayRT paN1 = toAADistribution(paOld);
		double[] pBackground={0.0839, 0.0126, 0.0593, 0.0681, 0.0409, 
	              0.0735, 0.0235, 0.0576, 0.0567, 0.0940, 
	              0.0167, 0.0427, 0.0463, 0.0375, 0.0517, 
	              0.0588, 0.0546, 0.0713, 0.0146, 0.0358};
		double[] paNew = new double[55];
		Rotamers rots = new Rotamers(4);
		for(int i=0;i<55;i++)
		{
			String rotString = rots.intToDunbrackCode(i);
			int aaType = ResName.triToInt(rotString.substring(0, 3));
			paNew[i] = paN1.getArray().get(i)/pBackground[aaType];
		}
		return new ProbabilityArrayRT(paNew);
	}
	
	public static double getDeltaS1(PDB pdb, StrideProtein sp, int resID, char natType, char mutType, ResPara rpr, Rotamers rots) throws IOException
	{
		int resIDSD = pdb.getFirstChain().getResidue(resID).getIDSD();
		
		StrideRes sr = sp.getStrideChains().get(0).getResidues().get(resIDSD);
		HashSet<String> banList = new HashSet<String>();
		ProbabilityArrayRT pa = similarSearch.FindSimilarResiduesRT.findPAofRes(sr, rpr, rots, banList);
		ProbabilityArrayRT Pb=ProbabilityArrayRT.getPB(4);
		pa.addPseudoCount(Pb);
		ProbabilityArrayRT pax = toAADistributionOverBackground(pa);
		int natRotType = rots.getAminoAcidRotamerID(ResName.sinToInt(natType)).get(0);
		int mutRotType = rots.getAminoAcidRotamerID(ResName.sinToInt(mutType)).get(0);
		double pNat = pax.getArray().get(natRotType);
		double pMut = pax.getArray().get(mutRotType);
		double deltaS1 = Math.log(pNat/pMut);
		return deltaS1;
	}
	
	public static double getDeltaS2(PDB wtPDB, PDB mutPDB, StrideProtein sp, int resID, PairPara ppr, Rotamers rots) throws IOException
	{
		int resIDSD = wtPDB.getFirstChain().getResidue(resID).getIDSD();
		ArrayList<RotPair> rps = wtPDB.getFirstChain().findContactRotPairs(8.0);
		ArrayList<RotPair> involvedPairs = new ArrayList<RotPair>();
		for(RotPair rp : rps)
		{
			if(rp.getResA().getIDSD() == resIDSD || rp.getResB().getIDSD() == resIDSD)
				involvedPairs.add(rp);
		}
		ArrayList<ScoreMatrixRT> smList = new ArrayList<ScoreMatrixRT>();
		ProbabilityArrayRT Pb=ProbabilityArrayRT.getPB(4);
		HashSet<String> banList = new HashSet<String>();
		for(RotPair rp : involvedPairs)
		{
			rp.addStrInfo(sp);
			ProbabilityMatrixRT pm = FindSimilarPairsRT.findPMofRotPair(rp, ppr, banList, rots);
			
			PairPara thisPpr = new PairPara(ppr.resParaA(),ppr.resParaB(),ppr.cutoff(),ppr.fqFloor());

			while(pm.maxFq()<thisPpr.fqFloor() && thisPpr.cutoff()<1.5)
			{
				thisPpr = thisPpr.demote();
				pm = FindSimilarPairsRT.findPMofRotPair(rp,thisPpr,banList,rots);
			}

			ProbabilityArrayRT paResA = FindSimilarResiduesRT.findPAofRes(rp.strResA, thisPpr.resParaA(),rots,banList);
			paResA.addPseudoCount(Pb);
			ProbabilityArrayRT paResB = FindSimilarResiduesRT.findPAofRes(rp.strResB, thisPpr.resParaB(),rots,banList);
			paResB.addPseudoCount(Pb);
			ProbabilityMatrixRT pmCross = paResA.cross(paResB);
			
			ScoreMatrixRT sm = new ScoreMatrixRT(pm,pmCross);
			
			sm.adjustment(pm);
			
			int seqDist = rp.seqDistance();
			if(seqDist < 3)
				sm.multiply(0.5);
			else if(seqDist < 5)
				sm.multiply(0.6);
			smList.add(sm);
			
			
		}
		String natRotSeq = wtPDB.getChains().get(0).getRotamerSeq();
		String mutRotSeq = mutPDB.getFirstChain().getRotamerSeq();
		double deltaS2 = getS2(mutRotSeq, smList, involvedPairs, rots) - getS2(natRotSeq, smList, involvedPairs, rots);
		return deltaS2;
	}
	
	
	public static double getDeltaS2(PDB pdb, StrideProtein sp, int resID, char natType, char mutType, PairPara ppr, Rotamers rots) throws IOException
	{
		int resIDSD = pdb.getFirstChain().getResidue(resID).getIDSD();
		ArrayList<RotPair> rps = pdb.getFirstChain().findContactRotPairs(8.0);
		ArrayList<RotPair> involvedPairs = new ArrayList<RotPair>();
		for(RotPair rp : rps)
		{
			if(rp.getResA().getIDSD() == resIDSD || rp.getResB().getIDSD() == resIDSD)
				involvedPairs.add(rp);
		}
		ArrayList<ScoreMatrixRT> smList = new ArrayList<ScoreMatrixRT>();
		ProbabilityArrayRT Pb=ProbabilityArrayRT.getPB(4);
		HashSet<String> banList = new HashSet<String>();
		
		for(RotPair rp : involvedPairs)
		{
			rp.addStrInfo(sp);
			ProbabilityMatrixRT pm = FindSimilarPairsRT.findPMofRotPair(rp, ppr, banList, rots);
			
			PairPara thisPpr = new PairPara(ppr.resParaA(),ppr.resParaB(),ppr.cutoff(),ppr.fqFloor());

			while(pm.maxFq()<thisPpr.fqFloor() && thisPpr.cutoff()<1.5)
			{
				thisPpr = thisPpr.demote();
				pm = FindSimilarPairsRT.findPMofRotPair(rp,thisPpr,banList,rots);
			}

			ProbabilityArrayRT paResA = FindSimilarResiduesRT.findPAofRes(rp.strResA, thisPpr.resParaA(),rots,banList);
			paResA.addPseudoCount(Pb);
			ProbabilityArrayRT paResB = FindSimilarResiduesRT.findPAofRes(rp.strResB, thisPpr.resParaB(),rots,banList);
			paResB.addPseudoCount(Pb);
			ProbabilityMatrixRT pmCross = paResA.cross(paResB);
			
			ScoreMatrixRT sm = new ScoreMatrixRT(pm,pmCross);
			
			sm.adjustment(pm);
			
			int seqDist = rp.seqDistance();
			if(seqDist < 3)
				sm.multiply(0.5);
			else if(seqDist < 5)
				sm.multiply(0.6);
			smList.add(sm);
		}
		
		String natRotSeq = pdb.getChains().get(0).getRotamerSeq();
		ArrayList<String> mutSeqs = mutate(natRotSeq, resIDSD, mutType);
		double natS2 = getS2(natRotSeq,smList,involvedPairs,rots);
		ArrayList<Double> mutS2s = new ArrayList<Double>();
		for(String mutSeq : mutSeqs)
		{
			mutS2s.add(getS2(mutSeq,smList,involvedPairs,rots));
		}
		double minMutS2 = min(mutS2s);
		return minMutS2 - natS2;
	}
	
	
	
	public static ArrayList<String> mutate(String natRotSeq, int site, char type) throws IOException
	{
		Rotamers rots = new Rotamers(4);
		ArrayList<Integer> rotIDs = rots.getAminoAcidRotamerID(ResName.sinToInt(type));
		ArrayList<String> mutSeqs = new ArrayList<String>();
		for(Integer rotID : rotIDs)
		{
			String duncode = rots.intToDunbrackCode(rotID);
			mutSeqs.add(natRotSeq.substring(0, site*7) + duncode + natRotSeq.substring(site*7+7, natRotSeq.length()));
		}
		return mutSeqs;
	}
	
	public static double getS2(String rotSeq, ArrayList<ScoreMatrixRT> smList, ArrayList<RotPair> rps, Rotamers rots)
	{
		double s2 = 0;
		
		for(int i=0;i<rps.size();i++)
		{
			RotPair rp = rps.get(i);
			int resAIDSD = rp.getResA().getIDSD();
			int resBIDSD = rp.getResB().getIDSD();
			int rotA = rots.dunbrackCodeToInt(rotSeq.substring(resAIDSD*7, resAIDSD*7+7));
			if(rotA == -1)
			{
				int resType = ResName.triToInt(rotSeq.substring(resAIDSD*7, resAIDSD*7+3));
				rotA = rots.getAminoAcidRotamerID(resType).get(2);
			}
			int rotB = rots.dunbrackCodeToInt(rotSeq.substring(resBIDSD*7, resBIDSD*7+7));
			if(rotB == -1)
			{
				int resType = ResName.triToInt(rotSeq.substring(resBIDSD*7, resBIDSD*7+3));
				rotB = rots.getAminoAcidRotamerID(resType).get(2);
			}
			s2 += smList.get(i).getScoreMatrix()[rotA][rotB];
		}
		return s2;
	}
	
	public static double min(ArrayList<Double> scores)
	{
		double minS = scores.get(0);
		for(Double s : scores)
		{
			if(s < minS)
				minS = s;
		}
		return minS;
	}
	
	public static ArrayList<String> getMutSites(String pdbID) throws IOException
	{
		ArrayList<String> lines = usr.FileToList.getList("/nfs/amino-home/pengx/mutationThermo/ddGExp");
		ArrayList<String> select = new ArrayList<String>();
		for(String line : lines)
		{
			if(line.startsWith(pdbID.toUpperCase())) select.add(line);
		}
		return select;
	}
	
	public static void scorePDB(String pdbID,  PrintStream out) throws IOException
	{
		ArrayList<String> lines = getMutSites(pdbID);
		String wtPDBFile = "/nfs/amino-home/pengx/mutationThermo/wtScwrlPDB/"+pdbID+".pdb";
		
		String strFile = "/nfs/amino-home/pengx/mutationThermo/wtSTR/"+pdbID+".str";
	
		PDB pdb = new PDB(wtPDBFile);
		StrideProtein sp = new StrideProtein(strFile);
		ResPara rpr = new ResPara(1, 4, 0);
		PairPara ppr = new PairPara(40, 0.5, 1, 2, 0);
		Rotamers rots = new Rotamers(4);
		for(String line : lines)
		{
			String[] spt = line.trim().split("\t");
			int resID = Integer.parseInt(spt[8]);
			char resTypeNat = spt[3].charAt(0);
			char resTypeMut = spt[4].charAt(0);
			String mutFileName = pdbID.substring(0, 4).toLowerCase() + pdbID.charAt(4) + resID + resTypeMut + ".pdb";
			PDB mutPDB = new PDB("/nfs/amino-home/pengx/mutationThermo/mutScwrlPDB/" + mutFileName);
			double deltaS1 = getDeltaS1(pdb, sp, resID, resTypeNat, resTypeMut, rpr, rots);
			double deltaS2 = getDeltaS2(pdb, mutPDB ,sp, resID, ppr, rots);
			double ddGExp = Double.parseDouble(spt[5])*(-1.0);
			
			out.printf("%s %c %3d %c %7.4f %7.4f %6.2f\n",pdbID,resTypeNat,resID,resTypeMut,deltaS1,deltaS2,ddGExp);
		}
	}
	
	public static void scorePDBNat(String pdbID,  PrintStream out) throws IOException
	{
		ArrayList<String> lines = getMutSites(pdbID);
		String wtPDBFile = "/nfs/amino-home/pengx/mutationThermo/nativePDBs/" + pdbID + "/"+pdbID+".pdb";
		
		String strFile = "/nfs/amino-home/pengx/mutationThermo/nativePDBs/" + pdbID + "/"+pdbID+".str";
	
		PDB pdb = new PDB(wtPDBFile);
		StrideProtein sp = new StrideProtein(strFile);
		ResPara rpr = new ResPara(1, 4, 0);
		PairPara ppr = new PairPara(40, 0.5, 1, 2, 0);
		Rotamers rots = new Rotamers(4);
		for(String line : lines)
		{
			String[] spt = line.trim().split("\t");
			int resID = Integer.parseInt(spt[2]);
			char resTypeNat = spt[3].charAt(0);
			char resTypeMut = spt[4].charAt(0);
			double deltaS1 = getDeltaS1(pdb, sp, resID, resTypeNat, resTypeMut, rpr, rots);
			double deltaS2 = getDeltaS2(pdb, sp, resID,resTypeNat,resTypeMut, ppr, rots);
			double ddGExp = Double.parseDouble(spt[5])*(-1.0);
			
			out.printf("%s %c %3d %c %7.4f %7.4f %6.2f\n",pdbID,resTypeNat,resID,resTypeMut,deltaS1,deltaS2,ddGExp);
		}
	}
	
	
	
	public static void getCorrelationValue() throws IOException
	{
		ArrayList<String> lines = usr.FileToList.getList("/nfs/amino-home/pengx/mutationThermo/results/S1-1-3-2-S2-40-0.5-1-2-2");
		ArrayList<Double> s2s = new ArrayList<Double>();
		ArrayList<Double> exps = new ArrayList<Double>();
		ArrayList<Double> s1s = new ArrayList<Double>();
		
		for(String s : lines)
		{
			String[] spt = s.split(" +");
			s2s.add(Double.parseDouble(spt[5]));
			s1s.add(Double.parseDouble(spt[4]));
			exps.add(Double.parseDouble(spt[6]));
		}
		for(double wt = 0.0;wt<3.0;wt+=0.1)
		{
			ArrayList<Double> s1s2 = new ArrayList<Double>();
			for(int i=0;i<s2s.size();i++)
				s1s2.add(s1s.get(i)*wt + s2s.get(i));
			System.out.printf("%3.1f %6.4f\n",wt,math.CorrelationCoefficient.correlation(s1s2, exps));
		}
		
	
	}
	
	public static void getCorrelation(String subList, String result) throws IOException
	{
		ArrayList<String> lines = usr.FileToList.getList(result);
		ArrayList<Double> s2s = new ArrayList<Double>();
		ArrayList<Double> exps = new ArrayList<Double>();
		ArrayList<Double> s1s = new ArrayList<Double>();
		
		ArrayList<String> subLines = usr.FileToList.getList(subList);
		HashSet<String> set = new HashSet<String>();
		for(String s : subLines)
			set.add(s);
		
		for(String s : lines)
		{
			if(!set.contains(s.substring(0,5))) continue;
			String[] spt = s.split(" +");
			s2s.add(Double.parseDouble(spt[5]));
			s1s.add(Double.parseDouble(spt[4]));
			exps.add(Double.parseDouble(spt[6]));
		}
		ArrayList<Double> s1s2 = new ArrayList<Double>();
		for(int i=0;i<s1s.size();i++)
			s1s2.add(s1s.get(i) + s2s.get(i));
		double coS1 = math.CorrelationCoefficient.correlation(s1s, exps);
		double coS2 = math.CorrelationCoefficient.correlation(s2s, exps);
		double coS12 = math.CorrelationCoefficient.correlation(s1s2, exps);
		
		System.out.printf("%6.4f %6.4f %6.4f\n",coS1,coS2,coS12);
		
	}
	
	
	public static void getCorrelationDiffWt(String subList) throws IOException
	{
		ArrayList<String> lines = usr.FileToList.getList("/nfs/amino-home/pengx/mutationThermo/results/S1-1-3-2-S2-40-0.5-1-2-2");
		ArrayList<Double> s2s = new ArrayList<Double>();
		ArrayList<Double> exps = new ArrayList<Double>();
		ArrayList<Double> s1s = new ArrayList<Double>();
		
		ArrayList<String> subLines = usr.FileToList.getList(subList);
		HashSet<String> set = new HashSet<String>();
		for(String s : subLines)
			set.add(s);
		
		for(String s : lines)
		{
			if(!set.contains(s.substring(0,5))) continue;
			String[] spt = s.split(" +");
			s2s.add(Double.parseDouble(spt[5]));
			s1s.add(Double.parseDouble(spt[4]));
			exps.add(Double.parseDouble(spt[6]));
		}
		for(double wt = 0.0;wt<3.0;wt+=0.1)
		{
			ArrayList<Double> s1s2 = new ArrayList<Double>();
			for(int i=0;i<s2s.size();i++)
				s1s2.add(s1s.get(i)*wt + s2s.get(i));
			System.out.printf("%3.1f %6.4f\n",wt,math.CorrelationCoefficient.correlation(s1s2, exps));
		}
		
	
	}
	
	public static void main(String[] args) throws IOException
	{
/*
		ArrayList<PairPara> pprList = new ArrayList<PairPara>();
			
		pprList.add(new PairPara(40, 0.5, 1, 2, 0));
		pprList.add(new PairPara(40, 0.7, 1, 2, 0));
		pprList.add(new PairPara(40, 1.0, 1, 2, 0));
		
			
		ArrayList<String> pdbList = TestAll.getTestPDBList();
		PrintStream out = new PrintStream("E:/projects/singleMutationThermo/s2-2-" + args[0] );
		for(String pdb: pdbList)
		{
			scorePDB(pdb, pprList.get(Integer.parseInt(args[0])),out);
		}
		out.close();
*/
	/*	
		int jobID = Integer.parseInt(args[0]);
		PrintStream out = new PrintStream("/nfs/amino-home/pengx/mutationThermo/sefDDG-"+jobID);
		ArrayList<String> list = usr.FileToList.getList("/nfs/amino-home/pengx/mutationThermo/wtScwList-"+jobID);
		for(String s : list)
		{
			scorePDB(s.substring(0,5),out);
		}
		out.close();
	*/
	//	getCorrelationValue();
		
		PrintStream out = new PrintStream("/nfs/amino-home/pengx/mutationThermo/sefDDG-natPDB-"+args[0]);
		scorePDBNat(args[0],out);
		out.close();
		
	//	getCorrelation(args[0],args[1]);
	}
}
