/******************************************************************************/
/*                                                                            */
/*                                xtd-seqali                                  */
/*                                                                            */
/* Computes the optimum alignment between two amino acid sequences (S1 & S2). */
/* The score of each pair (i,j) of positions takes into account not only      */
/* the similarity between the amino acids but also other features that        */
/* characterize those positions.                                              */
/*                                                                            */
/* In the current implementation, the score of each pair (i,j) of positions   */
/* is given by the secondary structures at positions i and j in S1 and S2,    */
/* respectively, and by the value of S1[i] at position j in the PSSM of S2.   */
/*                                                                            */
/* ARGUMENTS:                                                                 */
/* ---> S1:  path to the file containing amino acid sequence S1.              */
/*           The sequence is all on one line, starts at the start of the      */
/*           file, and is immediately followed by a line terminator.          */
/* ---> PF2: path to the file containing the PSSM of S2.                      */
/*           The PSSM can be output either by program makemat from the        */
/*           Impala suite, or by program mkprf (see notes).                   */
/* ---> SS1: path to the file containing the secondary structure of S1.       */
/*           The secondary structure is a horizontal sequence, with one       */
/*           character per residue position. It starts at the start of the    */
/*           file and is immediately followed by a line terminator.           */
/*           The i-th character represents the secondary structure at the     */
/*           i-th residue position (i=0,...,N-1, where N is the number of     */
/*           residues).            .                                          */
/* ---> SS2: path to the file containing the secondary structure of S2.       */
/*           The file format is the same as for SS1.                          */
/* ---> WSS: weight of the secondary structure in the alignment.              */
/*                                                                            */
/* NOTES:                                                                     */
/* - the PSSM output by program mkprf consists of a header line, N data lines,*/
/*   and a footer line.                                                       */
/*     The header line contains 21 columns, with one amino acid type per      */
/*   column, expressed in one-letter code. The last column, labeled '-'       */
/*   refers to the gap. The columns are separated by a certain number of      */
/*   blanks and the last column is immediately followd by a line terminator.  */
/*     The i-th data line contains the profile data for the i-th residue      */
/*   position, for i=0,...,N-1, where N is the number of residue positions.   */
/*   The j-th column is the score of the j-th amino acid type at that         */
/*   position, for j=0,...,19. The 20-th column is the score of the gap at    */
/*   that position. The 20-th column is immediately followed by a line        */
/*   terminator.                                                              */
/*     The footer line is identical to the header line. After its line        */
/*   terminator the file terminates.                                          */
/*                                                                            */
/******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "SeqAlign.h"
#include "getSeq2SA.h"
#include "getSeq2SS.h"
#include "getPhiPsi.h"
#include "utility.h"
#include <unistd.h>


namespace Text {
  int ncharLine(FILE* fp);
}

int main(int argc, char** argv) 
{
	char file[500],profile[500];
	double max;
	struct SequenceData dsInfo;

	if(argc!=9) {
		//printf("Usage:: ./xtdseqali seq.txt prf.txt ss.txt sa.txt phi-psi.txt working_path 1.0 1.0 1.0\n");
		printf("Usage:: ./xtdseqali seq.txt prf.txt ss.txt sa.txt phi-psi.txt 1.0 1.0 1.0\n");
//					    argv1   argv2   argv3  argv4  argv5       argv6        argv{7,8,9}
		exit(0);
	}
	
/////////////////////////////////////	
	char buf[1024]="";
   	int count=readlink("/proc/self/exe", buf, 1024);
   	if(count<0 || count>=1024)
   	{
      		printf ("Error in finding executive directory!\n");
      		exit(0); 
   	}
   	for(int i=1023; i>=0; i--)
   	{
      		if(buf[i]=='/') break;
      		else buf[i]='\0';
   	}
   	char param_dir[1024];
   	sprintf(param_dir,"%s/../param/",buf);


/* Read input sequence 
	Input: Filename containing the sequence (Argument 1)
	Output: seq will store seqeunce and seqlen will contain length of sequence
*/
	sprintf(file,"./%s", argv[1]);		// sequence file

	FILE* fp = fopen(file, "r");
//	const int seqlen = Text::ncharLine(fp);
/////////////////////////////////////////////////////////////////////////////////
//printf("FILE: %s\n", file);
	dsInfo.len= Text::ncharLine(fp);

/////////////////////////////////////////////////////////////////////////////////

	dsInfo.seq = new char[dsInfo.len+1];
	fseek(fp, 0, SEEK_SET);
	fscanf(fp, "%s", dsInfo.seq);
	fclose(fp);

/* Secondary structure prediction from single sequence 
	Input: sequence and current path containing weight and propensity files
	Output: predicted secondary structure in ss1 
*/
	dsInfo.ss1 = new char[dsInfo.len+1];
	SSPrediction::getSeq2SS SS(dsInfo.ss1,dsInfo.seq, param_dir);

/* Solvent accessibility prediction from single sequence 
	Input: sequence, predicted secondary structure and current path containing weight and propensity files	
	Output: predicted solvent accessibility
*/
	dsInfo.sa1 = new char[dsInfo.len+1];
	SAPrediction::getSeq2SA SA(dsInfo.sa1,dsInfo.seq,dsInfo.ss1, param_dir);

/* Phi Psi angle prediction from single sequence */
	PhiPsiPrediction::getPhiPsi(dsInfo, param_dir);

	// read the secondary structure of target
	dsInfo.ss2 = new char[dsInfo.len+1];
	sprintf(file,"./%s",argv[3]);		// target secondary structure file
	fp = fopen(file, "r");
	fscanf(fp, "%s", dsInfo.ss2);
	fclose(fp);

	// read the solvent accessibility of target
	dsInfo.sa2 = new char[dsInfo.len+1];
	sprintf(file,"./%s",argv[4]);		// target secondary structure file
	fp = fopen(file, "r");
	fscanf(fp, "%s", dsInfo.sa2);
	fclose(fp);

	// weight functions
	float wss = atof(argv[6]);
	float wsa = atof(argv[7]);
	float wang= atof(argv[8]);

	wsa /= (dsInfo.len*1.0);		// weight for solvent accessibility
	wss /= (dsInfo.len*1.0);		// weight for secondary structure
	wang/= (dsInfo.len*180.0);		// weight for phi-psi prediction

/* NW based alignment using profile, secondary structure, solvent accessibility and phi-psi angle scoring */

	sprintf(profile,"./%s",argv[2]);		// profile
	CharSeq::SeqAlign::SeqAlign sa(dsInfo, profile, "./", wss, wsa, wang);
	max=sa.printMtxInfo();
	printf("%f",max);

	delete[] dsInfo.ss1;	delete[] dsInfo.ss2;
	delete[] dsInfo.sa1;	delete[] dsInfo.sa2;
	delete[] dsInfo.seq;

	return 1;
}
